/*
 *
 * Copyright (C) 2004 Alexander Yip (yipal@lcs.mit.edu)
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <sys/types.h>
#include "arpc.h"
#include "dbfe.h"
#include "bigint.h"
#include "str.h"
#include "parseopt.h"
#include "chord_types.h"
#include "rxx.h"
#include "misc_utils.h"
#include "id_utils.h"
#include <ihash.h>
#include "dhash_blockstore.h"
#include "dhash_types.h"
#include "audit_prot.h"
#include "merkle_misc.h"

str lslogfile;
str dbdir;
ptr<dbfe> project_to_blocks;
ptr<dbfe> block_db;
ptr<dbfe> keyhash_db;

unsigned int rpc_calls;
ptr<aclnt> gwclnt;
ptr<dhash_blockstore> storage;

struct block_entry {
  chordID key;
  ihash_entry<block_entry> link;
  block_entry(chordID k) {key = k;}
};

ihash<chordID, block_entry, &block_entry::key, 
  &block_entry::link, hashID> legal_blocks;


ptr<dhash_blockstore>
open_db(str dbfilename)
{
  ptr<dhash_blockstore> storage = New refcounted<dhash_blockstore>();
  storage->opendb(dbfilename, false, 30);
  return storage;
}

rxx blocklist_rxx ("(^[0-9a-zA-Z+-]+)$");
vec<bigint>
parse_blockids(str filecontents)
{
  int r, i;
  vec<str> lines;
  vec<bigint> ids;
  r = split(&lines, "\n", filecontents);
  
  for(i=0; i<r; i++) {
    if (blocklist_rxx.match(lines[i])) {
      warn << i << ": " << blocklist_rxx[1] << "\n";
      ids.push_back(dearmor64A_bigint(blocklist_rxx[1]));
    } else
      warn << "error parsing blocklist line " << 1+i << "\n";
  }
  return ids;
}

int 
retrieve_blocklist(chordID projectid)
{
  unsigned int i;
  str rawblock_ids;
  vec<bigint> block_ids;
  struct block_entry *b;
  
  //call lslog, read/parse the output
  strbuf sb = strbuf() << "lslog " << armor64A(projectid) << " > " << lslogfile;
  if (system (str(sb).cstr())){
    warn << "Error auditing project " << armor64A(projectid) << ". Skipping!\n";
    return -1;
  }

  if (!(rawblock_ids = file2str(lslogfile))){
    warn << "Could not read temp file for project " 
	 << armor64A(projectid) << ". Skipping!\n";
    return -1;
  }

  block_ids = parse_blockids(rawblock_ids);
  for(i=0; i<block_ids.size(); i++) {
    b = New block_entry(block_ids[i]);
    legal_blocks.insert(b);
  }

  return 0;
}

void rpc_return(ptr<audit_removenotify_arg> arg, dhash_ctype ctype, 
		auditstat *res, clnt_stat err) {
  warn << "Returned from rpc call " << err << " " << *res << "\n";

  if (!err && !(*res)) { // Only remove the block after the RPC returns successfully
    switch (ctype) {
    case DHASH_CONTENTHASH:
      storage->remove_chblock(id2dbrec(arg->removedblockid));
      break;
    case DHASH_KEYHASH:
      storage->remove_pkblock(id2dbrec(arg->removedblockid));
      break;
    default:
      warn << "Unknown block type!\n";
      break;
    }
  }
  
  delete  res;
  rpc_calls--;
  if (rpc_calls == 0) 
    exit(0);
}

void initialize_rpc(str sockname) {
  int fd = unixsocket_connect(sockname);
  if (fd < 0) {
    fatal ("auditutil: Error connecting to %s: %s\n",
	   sockname.cstr (), strerror (errno));
  }
  gwclnt = aclnt::alloc
    (axprt_unix::alloc (fd, 1024*1025), audit_program_1);
}

void remove_block(bigint key, dhash_ctype ctype) {
  auditstat *res = New auditstat;
  ptr<audit_removenotify_arg> arg = New refcounted<audit_removenotify_arg>();
  arg->removedblockid = key;
  
  switch (ctype) {
  case DHASH_CONTENTHASH:
    gwclnt->call (AUDITPROC_REMOVE_CH, arg, res, 
		  wrap (&rpc_return, arg, ctype, res));
    rpc_calls++;
    break;
  case DHASH_KEYHASH:
    gwclnt->call (AUDITPROC_REMOVE_PK, arg, res, 
		  wrap (&rpc_return, arg, ctype, res));
    rpc_calls++;
    break;
  default:
    fatal << "unknown ctype\n";
  }
}



void audit(chordID projectid, str auditsock)
{ 
  unsigned int i;
  warn << "auditing: " << armor64A(projectid) << " (0x" << projectid << ")\n";
  
  if (retrieve_blocklist(projectid)) {
    warn << "Could not retrieve project's block list. Bailing\n";
    exit(1);
  }

  rpc_calls = 0;
  storage = open_db(dbdir);
  initialize_rpc(auditsock);

  vec<ptr<dbrec> > stored_keys;

  if (storage->getpkblockids(&stored_keys, &projectid)) 
    fatal << "could not get pk blocks from database.\n";
  for(i=0; i<stored_keys.size(); i++) {
    if (legal_blocks[ dbrec2id(stored_keys[i]) ]) {
      // Block is ok, leave it
      warn << "block " << armor64A(dbrec2str(stored_keys[i])) 
	   << " is legal. Keeping\n";
    } else {
      // Block is stray, delete it.
      warn << "block " << armor64A(dbrec2str(stored_keys[i])) 
	   << " is illegal. Removing\n";
      remove_block(dbrec2id(stored_keys[i]), DHASH_KEYHASH);
    }
  }

  if (storage->getchblockids(&stored_keys, &projectid)) 
    fatal << "could not get ch blocks from database.\n";
  for(i=0; i<stored_keys.size(); i++) {
    if (legal_blocks[ dbrec2id(stored_keys[i]) ]) {
      // Block is ok, leave it
      warn << "block " << armor64A(dbrec2str(stored_keys[i])) 
	   << " is legal. Keeping\n";
    } else {
      // Block is stray, delete it.
      warn << "block " << armor64A(dbrec2str(stored_keys[i])) 
	   << " is illegal. Removing\n";
      remove_block(dbrec2id(stored_keys[i]), DHASH_CONTENTHASH);
    }
  }

  if (rpc_calls)
    amain();
}

void
usage() 
{
  warn << "usage: aquaaudit -p projkey(armor64A encoded) -S audit_socket -d db_directory [-l tempfile]\n";
  exit(1);
}

int
main (int argc, char **argv)
{
  bool projkeyinit = false;
  char ch;
  lslogfile = "/tmp/audit-output";
  bigint projkey;
  str auditsock = "";
  dbdir = "";

  while ((ch = getopt (argc, argv, "d:l:p:S:"))!=-1)
    switch (ch) {
    case 'l':
      lslogfile = optarg;
      break;
    case 'd':
      dbdir = optarg;
      break;
    case 'p':
      projkey = dearmor64A_bigint(optarg);
      projkeyinit = true;
      break;
    case 'S':
      auditsock = optarg;
      break;
    default:
      usage ();
      break;
    }
  
  if (!projkey)
    usage();

  if (auditsock == "") {
    warn << "You must specify an audit socket.\n";
    usage();
  }

  if (dbdir == "") {
    warn << "You must specify a database directory.\n";
    usage();
  }

  audit(projkey, auditsock);
  return 0;
}


