/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include "sfsmisc.h"
#include "arpc.h"
#include "async.h"
#include "str.h"
#include "parseopt.h"
#include "refcnt.h"
#include "bigint.h"
#include "dbfe.h"
#include "dhashgateway_prot.h"
#include "aquasingle.h"

static str aquasingle_socket;
static str db_dir = ".";
static int myport = 20000;

ptr<aclnt> local = 0;
ptr<dbfe> cache_db = 0;

static void
sync_diskcache_cb (ptr<dbfe> db)
{
  db->sync ();
  delaycb (30, wrap (&sync_diskcache_cb, db));
}

static void
client_accept (int fd)
{
  if (fd < 0)
    fatal ("EOF\n");
  assert (cache_db);
  ref<axprt_stream> x = axprt_stream::alloc (fd, 1024*1025);
  vNew refcounted<aquasinglesrv> (x);
}

static void
client_accept_socket (int lfd)
{
  sockaddr_un sun;
  bzero (&sun, sizeof (sun));
  socklen_t sunlen = sizeof (sun);
  int fd = accept (lfd, reinterpret_cast<sockaddr *> (&sun), &sunlen);
  if (fd >= 0)
    client_accept (fd);
}

static void
client_listen (int fd)
{
  if (listen (fd, 5) < 0) {
    fatal ("Error from listen: %m\n");
    close (fd);
  }
  else
    fdcb (fd, selread, wrap (client_accept_socket, fd));
}

EXITFN (cleanup);

static void
cleanup ()
{
  unlink (aquasingle_socket);
}

static void
startclntd ()
{
  unlink (aquasingle_socket);
  int clntfd = unixsocket (aquasingle_socket);
  if (clntfd < 0) 
    fatal << "Error creating client socket (UNIX)" << strerror (errno) << "\n";
  client_listen (clntfd);
  
  int port = myport;
  int tcp_clntfd = inetsocket (SOCK_STREAM, port);
  if (tcp_clntfd < 0)
    fatal << "Error creating client socket (TCP) " << strerror(errno) << "\n";
  client_listen (tcp_clntfd);
}

static void
open_worker (ptr<dbfe> mydb, str name, dbOptions opts, str desc)
{
  if (int err = mydb->opendb (const_cast <char *> (name.cstr ()), opts)) {
    warn << desc << ": " << name <<"\n";
    warn << "open returned: " << strerror (err) << "\n";
    exit (-1);
  }
}

static void
initialize_db ()
{
  dbOptions opts;
  opts.addOption ("opt_async", 1);
  opts.addOption ("opt_cachesize", 1000);
  opts.addOption ("opt_nodesize", 4096);

  cache_db = New refcounted<dbfe> ();
  str cdbs = strbuf () << db_dir << "/db.c";
  open_worker (cache_db, cdbs, opts, "cache db file");
  delaycb (30, wrap (&sync_diskcache_cb, cache_db));
  startclntd ();
}

static void
usage ()
{
  warnx << "Usage: " << progname << " "
        << "[-d <db dir>] "
	<< "[-S <socket>] "
	<< "[-p <port>]\n";
  exit (1);
}

int
main (int argc, char **argv)
{
  setprogname (argv[0]);
  mp_clearscrub ();
  random_init ();
  aquasingle_socket = "/tmp/aquasingle-sock";
  int ch;

  while ((ch = getopt (argc, argv, "d:S:p:"))!=-1)
    switch (ch) {
    case 'p':
      myport = atoi (optarg);
      break;
    case 'd':
      db_dir = optarg;
      break;
    case 'S':
      aquasingle_socket = optarg;
      break;
    default:
      usage ();
      break;
    }
  initialize_db();
  amain ();
}


