/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "stdinc.h"
#include "pack.h"
#include "cfsw_record.h"
 
bool
cfsw_sattr::parse (xin *in)
{
  if (in->rint (valid_size) &&
      in->rquad (size) && 
      in->rint (valid_mode) &&
      in->rint ((int&) mode))
    return true;
  return false;
}

attr_delta::attr_delta (bigint object, cfsw_sattr attr)
  : delta (cfsw_delta::T_ATTR),
    _object (object), _attr (attr)
{
}

ptr<attr_delta>
attr_delta::alloc (str s)
{
  xin in (s);
  ptr<attr_delta> x = New refcounted<attr_delta> ();
  if (!x->parse_meta (&in))
    return 0;

  bigint o;
  cfsw_sattr a;
  if (in.rbigint (o) && a.parse (&in)) {
    x->_object = o;
    x->_attr = a;
    return x;
  }
  else
    return 0;
}

str
attr_delta::serialize () const
{
  strbuf sb;

  serialize_meta (sb);
  sbigint (sb, _object);
  _attr.serialize (sb);
  return sb;
}


inode_delta::inode_delta (cfsw_ftype object_type, bigint object)
  : delta (cfsw_delta::T_INODE),
    _object_type (object_type), _object (object), _attr ()
{
}

inode_delta::inode_delta (cfsw_ftype object_type, bigint object,
                                    cfsw_sattr sa)
  : delta (cfsw_delta::T_INODE),
    _object_type (object_type), _object (object), _attr (sa)
{
}

inode_delta::inode_delta (cfsw_ftype object_type)
  : delta (cfsw_delta::T_INODE),
    _object_type (object_type), _attr ()
{
  // random new i-number.
  char xxb[20];
  rnd.getbytes(xxb, 20);
  mpz_set_rawmag_be(&_object, xxb, 20);
}

inode_delta::inode_delta (cfsw_ftype object_type, cfsw_sattr sa)
  : delta (cfsw_delta::T_INODE),
    _object_type (object_type), _attr (sa)
{
  // random new i-number.
  char xxb[20];
  rnd.getbytes(xxb, 20);
  mpz_set_rawmag_be(&_object, xxb, 20);
}

ptr<inode_delta>
inode_delta::alloc (str s)
{
  xin in (s);
  ptr<inode_delta> x = New refcounted<inode_delta> ();
  if (!x->parse_meta (&in))
    return 0;

  cfsw_sattr a;
  bigint o;
  int i;
  if (in.rint (i) &&
      in.rbigint (o) &&
      a.parse (&in)) {
    x->_object_type = (cfsw_ftype)i;
    x->_object = o;
    x->_attr = a;
    return x;
  }
  else
    return 0;
}

str
inode_delta::serialize () const
{
  strbuf sb;
  serialize_meta (sb);
  sint(sb, _object_type);
  sbigint (sb, _object);
  _attr.serialize (sb);
  return sb;
}


dirop_delta::dirop_delta
  (int rtype, bigint object, cfsw_ftype object_type,
   bigint dir, str name)
  : delta (rtype),
    _object (object), _object_type (object_type), _dir(dir), _name(name)
{
}

link_delta::link_delta
  (bigint object, cfsw_ftype object_type, bigint dir, str name)
  : dirop_delta (cfsw_delta::T_LINK, object,
                      object_type, dir, name)
{
}

ptr<link_delta>
link_delta::alloc (str s)
{
  xin in (s);
  ptr<link_delta> x = New refcounted<link_delta> ();
  if (!x->parse_meta (&in))
    return 0;

  bigint o;
  bigint d;
  str n;
  int i;

  if (in.rbigint (o) &&
      in.rint (i) &&
      in.rbigint (d) &&
      in.rstr (n)) {
    x->_object = o;
    x->_object_type = (cfsw_ftype)i;
    x->_dir = d;
    x->_name = n;
    return x;
  }
  else
    return 0;
}

str
link_delta::serialize () const
{
  strbuf sb;
  serialize_meta (sb);
  sbigint (sb, _object);
  sint(sb, (int) _object_type);
  sbigint (sb, _dir);
  sstr(sb, _name);
  return sb;
}

unlink_delta::unlink_delta
  (bigint object, cfsw_ftype object_type, bigint dir, str name)
  : dirop_delta (cfsw_delta::T_UNLINK, object,
                      object_type, dir, name)
{
}

ptr<unlink_delta>
unlink_delta::alloc (str s)
{
  xin in (s);
  ptr<unlink_delta> x = New refcounted<unlink_delta> ();
  if (!x->parse_meta (&in))
    return 0;

  bigint o;
  int i;
  bigint d;
  str n;

  if (in.rbigint (o) &&
      in.rint (i) &&
      in.rbigint (d) &&
      in.rstr (n)) {
    x->_object = o;
    x->_object_type = (cfsw_ftype)i;
    x->_dir = d;
    x->_name = n;
    return x;
  }
  else
    return 0;
}

str
unlink_delta::serialize () const
{
  strbuf sb;
  serialize_meta (sb);
  sbigint (sb, _object);
  sint(sb, (int) _object_type);
  sbigint (sb, _dir);
  sstr(sb, _name);
  return sb;
}

rename_delta::rename_delta
  (bigint object, cfsw_ftype object_type,
   bigint from_dir, str from_name,
   bigint dir, str name, bigint target)

  : dirop_delta (cfsw_delta::T_RENAME, object,
                      object_type, dir, name),
    _from_dir (from_dir), _from_name (from_name),
    _target_exists (true), _target (target)
{
}

rename_delta::rename_delta
  (bigint object, cfsw_ftype object_type,
   bigint from_dir, str from_name,
   bigint dir, str name)
  : dirop_delta (cfsw_delta::T_RENAME, object,
                      object_type, dir, name),
    _from_dir (from_dir), _from_name (from_name), _target_exists (false)
{
}

ptr<rename_delta>
rename_delta::alloc (str s)
{
  xin in (s);
  ptr<rename_delta> x = New refcounted<rename_delta> ();
  if (!x->parse_meta (&in))
    return 0;

  bigint o, fd, d, t;
  str fn, n;
  int target;
  int i;

  if (in.rbigint (o) &&
      in.rint (i) &&
      in.rbigint (fd) &&
      in.rstr (fn) &&
      in.rbigint (d) &&
      in.rstr (n) &&
      in.rint (target) &&
      in.rbigint (t)) {
    if (target) {
      x->_object = o;
      x->_object_type = (cfsw_ftype)i;
      x->_dir = d;
      x->_name = n;
      x->_from_dir = fd;
      x->_from_name = fn;
      x->_target_exists = true;
      x->_target = t;
      return x;
    }
    else {
      x->_object = o;
      x->_object_type = (cfsw_ftype)i;
      x->_dir = d;
      x->_name = n;
      x->_from_dir = fd;
      x->_from_name = fn;
      x->_target_exists = false;
      return x;
    }
  }
  else
    return 0;
}

str
rename_delta::serialize () const
{
  strbuf sb;
  serialize_meta (sb);
  sbigint (sb, _object);
  sint(sb, (int) _object_type);
  sbigint (sb, _from_dir);
  sstr(sb, _from_name);
  sbigint (sb, _dir);
  sstr(sb, _name);
  if (_target_exists)
    sint(sb, 1);
  else
    sint(sb, 0);
  sbigint (sb, _target);
  return sb;
}


symlink_delta::symlink_delta (bigint file, str dest)
  : delta (cfsw_delta::T_SYMLINK), _file (file), _dest (dest)
{
}

ptr<symlink_delta>
symlink_delta::alloc (str s)
{
  xin in (s);
  ptr<symlink_delta> x = New refcounted<symlink_delta> ();
  if (!x->parse_meta (&in))
    return 0;

  bigint f;
  str d;

  if (in.rbigint (f) &&
      in.rstr (d)) {
    x->_file = f;
    x->_dest = d;
    return x;
  }
  else
    return 0;
}

str
symlink_delta::serialize () const
{
  strbuf sb;
  serialize_meta (sb);
  sbigint (sb, _file);
  sstr (sb, _dest);
  return sb;
}

patch_delta::patch_delta (bigint file, const str& patch)
  : delta (cfsw_delta::T_PATCH), _file (file), _patch (patch)
{
}

ptr<patch_delta>
patch_delta::alloc (str s)
{
  xin in (s);
  ptr<patch_delta> x = New refcounted<patch_delta> ();
  if (!x->parse_meta (&in))
    return 0;
  if (in.rbigint (x->_file) &&
      in.rstr (x->_patch))
    return x;
  else
    return 0;
}

str
patch_delta::serialize () const
{
  strbuf sb;
  serialize_meta (sb);
  sbigint (sb, _file);
  sstr (sb, _patch);
  return sb;
}

write_delta::write_delta 
  (bigint file, const vec<bigint> &d,
   bigint sid, bigint did, bigint tid, uint64 size)
  : delta (cfsw_delta::T_WRITE)
{
  _file = file;
  _direct = d;
  _indirect = sid;
  _double_indirect = did;
  _triple_indirect = tid;
  _size = size;
}

ptr<write_delta>
write_delta::alloc (str s)
{
  xin in (s);
  ptr<write_delta> x = New refcounted<write_delta> ();
  if (!x->parse_meta (&in))
    return 0;

  bigint f;
  vec<bigint> d;
  bigint sid;
  bigint did;
  bigint tid;
  uint64 size;
  int n;

  if (in.rbigint (f) && in.rint (n)) {
    for (int i=0; i<n; i++) {
      bigint x;
      if (in.rbigint (x))
        d.push_back (x);
      else
	goto bad;
    }

    if (in.rbigint (sid) &&
        in.rbigint (did) &&
        in.rbigint (tid) &&
        in.rquad   (size)) {
      x->_file = f;
      x->_direct = d;
      x->_indirect = sid;
      x->_double_indirect = did;
      x->_triple_indirect = tid;
      x->_size = size;
      return x;
    }
    else
      goto bad;
  }
bad:
  return 0;
}

str
write_delta::serialize () const
{
  strbuf sb;
  serialize_meta (sb);
  sbigint (sb, _file);
  
  sint (sb, _direct.size ());
  for (unsigned i=0; i<_direct.size (); i++)
    sbigint (sb, _direct [i]);
  sbigint (sb, _indirect);
  sbigint (sb, _double_indirect);
  sbigint (sb, _triple_indirect);
  squad   (sb, _size);

  return sb;
}

tag_delta::tag_delta (bigint commit, const str& tag)
  : delta (cfsw_delta::T_TAG), _commit (commit), _tag (tag)
{
}

ptr<tag_delta>
tag_delta::alloc (str s)
{
  xin in (s);
  ptr<tag_delta> x = New refcounted<tag_delta> ();
  if (!x->parse_meta (&in))
    return 0;
  if (in.rbigint (x->_commit) &&
      in.rstr (x->_tag))
    return x;
  else
    return 0;
}

str
tag_delta::serialize () const
{
  strbuf sb;
  serialize_meta (sb);
  sbigint (sb, _commit);
  sstr (sb, _tag);
  return sb;
}

deltap
cfsw_deltap_parse (str s)
{
  xin in (s);
  unsigned type;
  if (!delta::parse_type (&in, type))
    return 0;
 
  deltap r;
  if (type == cfsw_delta::T_INODE)
    r = inode_delta::alloc (s);
  else if (type == cfsw_delta::T_LINK)
    r = link_delta::alloc (s);
  else if (type == cfsw_delta::T_UNLINK)
    r = unlink_delta::alloc (s);
  else if (type == cfsw_delta::T_RENAME)
    r = rename_delta::alloc (s);
  else if (type == cfsw_delta::T_PATCH)
    r = patch_delta::alloc (s);
  else if (type == cfsw_delta::T_WRITE)
    r = write_delta::alloc (s);
  else if (type == cfsw_delta::T_SYMLINK)
    r = symlink_delta::alloc (s);
  else if (type == cfsw_delta::T_ATTR)
    r = attr_delta::alloc (s);
  else if (type == cfsw_delta::T_TAG)
    r = tag_delta::alloc (s);

  if (r)
    return r;
  else
    return 0;
}
 
