/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "gcc_compat.h"
#include <sys/time.h>
#include "stdinc.h"
#include "parseopt.h"
#include "rxx.h"

#include "ns.h"
#include "vervec.h"
#include "traverse.h"
#include "projinterface.h"
#include "cfsw_prot.h"
#include "record.h"
#include "cfsw_record.h"
#include "snapshot.h"
#include "cfs.h"

int ns_debug = 0;
int cfs_debug = 0;

ptr<ns> nsobj = 0;
ptr<projinterface> _projinf;
vec<crec> _commits;
vec<bigint> _commith;

static bool verbose = false;
static int vfd = -1;
static comvec lastv;
static int work = 0;

vec<deltap> _deltas;
vec<bigint> _hashes;
unsigned _ndeltas = 0;

static void
finalize ()
{
  if (work == 0) {
    // this is the program exit point
    if (vfd >= 0) {
      strbuf b;
      lastv.serialize (b);
      str x = b;
      unsigned n = 0;
      unsigned r = 0;
      while (n < x.len () &&
	     (r = write (vfd, x.cstr ()+n, x.len ()-n)) > 0)
        n += r;
      close (vfd);
    }
    exit (0);
  }
}

extern void iterate_write
(deltap r, callback<void,bigint,cfsrw_status>::ref cb);

static void
dump_write_cb (bigint n, cfsrw_status status)
{
  if (!status) {
    strbuf out;
    out << armor64A (n) << "\n";
    out.tosuio ()->output (1);
  }
  else {
    if (status != CFSRWERR_NOMORELOG) {
      warn << "iterate write failed!\n";
      exit (1);
    }
    work--;
    finalize ();
  }
}

static void
dump_write (deltap r)
{
  work++;
  iterate_write (r, wrap (dump_write_cb));
}

static void
got_record (deltap r, bigint hash)
{
  assert (r);

  strbuf out;
  
  if (!verbose) {
    out << armor64A (hash) << "\n";
    if (r->rtype () == cfsw_delta::T_WRITE)
      dump_write (r);
  }
  else {
    out << "\n              "
        << " hash " << armor64A (hash) << "\n"
        << "              ";

    switch (r->rtype ()) {
      case cfsw_delta::T_INODE:
          {
	    ptr<inode_delta> l =
	      *reinterpret_cast<ptr<inode_delta>*> (&r);
	    switch (l->object_type ()) {
	      case CFSW_REG:
	        out << "  reg ";
	        break;
	      case CFSW_DIR:
	        out << "  dir ";
	        break;
	      case CFSW_LNK:
	        out << "  lnk ";
	        break;
	      default:
	        break;
	    }
	    out << armor64A (l->object ()) << "\n";
          }
          break;
      case cfsw_delta::T_RENAME:
	  {
	    out << "rname ";
            rename_delta *l = r->TEMPLATE record<rename_delta> ();
	    out << armor64A (l->object ()) << "\n                 in ";
	    out << armor64A (l->dir ())    << "\n                 as "
	        << l->name ()   << "\n";
	    out << "                 was " << l->from_name () << "\n";
	  }
          break;
      case cfsw_delta::T_LINK:
	  {
	    out << " link ";
            link_delta *l = r->TEMPLATE record<link_delta> ();
	    out << armor64A (l->object ()) << "\n                 in ";
	    out << armor64A (l->dir ())    << "\n                 as "
	        << l->name ()   << "\n";
	  }
          break;
      case cfsw_delta::T_UNLINK:
	  {
	    out << "ulink ";
            unlink_delta *l = r->TEMPLATE record<unlink_delta> ();
	    out << armor64A (l->object ()) << "\n                 in ";
	    out << armor64A (l->dir ())    << "\n                 as "
	        << l->name ()   << "\n";
	  }
          break;
      case cfsw_delta::T_SYMLINK:
	  {
            out << "slink ";
            symlink_delta *l = r->TEMPLATE record<symlink_delta> ();
	    out << armor64A (l->file ())   << "\n                 to "
	        << l->dest () << "\n";
	  }
          break;
      case cfsw_delta::T_TAG:
	  {
            out << "  tag ";
            tag_delta *l = r->TEMPLATE record<tag_delta> ();
	    out << armor64A (l->commit ()) << "\n               with "
	        << l->tag () << "\n";
	  }
          break;
      case cfsw_delta::T_PATCH:
	  {
            out << "patch ";
            patch_delta *l = r->TEMPLATE record<patch_delta> ();
	    out << armor64A (l->file ()) << "\n";
	  }
          break;
      case cfsw_delta::T_WRITE:
	  {
            out << "write ";
            write_delta *l = r->TEMPLATE record<write_delta> ();
	    out << armor64A (l->file ()) << "\n               size ";
	    out << l->size () << " bytes\n";
	  }
          break;
      case cfsw_delta::T_ATTR:
	  {
	    out << " attr ";
            attr_delta *l = r->TEMPLATE record<attr_delta> ();
	    out << armor64A (l->object ()) << "\n               size ";
	    if (l->attr ()->valid_size)
	      out << l->attr ()->size << " bytes\n";
	    else
	      out << " not set\n";
	  }
          break;
      default:
          break;
    }
  }

  out.tosuio ()->output (1);
}

static void get_deltas ();

static void
fetch_deltap_cb (unsigned i, ptr<ns_block> b)
{
  if (!b) {
    warn << "cannot load record " << armor64A (_hashes [i]) << "\n";
    exit (1);
  }

  str s (b->data, b->len);
  deltap r = cfsw_deltap_parse (s);
  if (!r) {
    warn << "cannot parse record " << armor64A (_hashes [i]) << "\n";
    exit (1);
  }
  _deltas [i] = r;
  _ndeltas++;

  if (_ndeltas == _deltas.size ()) {
    for (unsigned i=0; i<_deltas.size (); i++)
      got_record (_deltas [i], _hashes [i]);
    get_deltas ();
  }
}

static void
get_deltas_cb (crec c, bigint n, bool failed)
{
  if (failed) {
    warn << "cannot load record\n";
    exit (1);
  }

  unsigned i = _hashes.size ();
  _hashes.push_back (n);
  nsobj->retrieve_block (n, wrap (fetch_deltap_cb, i));
}

static void
get_ind_cb (bigint n)
{
  if (!verbose) {
    strbuf out;
    out << armor64A (n) << "\n";
    out.tosuio ()->output (1);
  }
}

static void
get_deltas ()
{
  if (_commits.empty ()) {
    assert (work > 0);
    work--;
    finalize ();
    return;
  }

  _deltas.clear ();
  _hashes.clear ();
  _ndeltas = 0;

  crec c = _commits.pop_back ();
  bigint n = _commith.pop_back ();

  strbuf out;
  if (verbose) {
    commit_info inf = c->depends();

    out << "\n"
        << "Commit " << armor64A (n) << "\n"
	<< "name: " << c->proposed_name () << "\n"
	<< "btag: " << c->btag () << "\n" 
        << "by: " << armor64A (c->head ()) << "\n"
        << "d.rptr: " << armor64A (inf.rptr) << "\n"
        << "d.vptr: " << armor64A (inf.vptr) << "\n"
        << "d.head: " << armor64A (inf.head) << "\n"
	<< "flags: " << c->flags () << "\n"
	<< "swver: " << c->swver() << "\n"
        << myctime (c->time ()) << "\n";
  }
  else
    out << armor64A (n) << "\n";
  out.tosuio ()->output (1);

  if (c->ndeltas () == 0)
    get_deltas ();
  else {
    _deltas.setsize (c->ndeltas ());
    get_deltas (c, wrap (get_ind_cb), wrap (get_deltas_cb, c));
  }
}

static void
fetch_crec_cb (iterinfo *i)
{
  if (i->status) {
    warn << "status " << i->status << "\n";
    exit (1);
  }

  if (!i->finished) {
    _commits.push_back (i->commit);
    _commith.push_back (i->hash);
  }
  else {
    work++;
    get_deltas ();
  }
}

static bool
commit_checker (bigint c)
{
  return false;
}

static void
got_projinf (bigint vid, ptr<head_state> v, log_status status)
{ 
  if (status) {
    warn << "cannot load project information\n";
    exit (1);
  }

  strbuf out;
  if (verbose)
    out << "Project ID: ";
  out << armor64A (vid);
  if (verbose)
    out << " " << _projinf->get_projblk_version();
  out << "\n";

  for (unsigned i=0; i<_projinf->nusers(); i++) {
    if (verbose)
      out << "Member block (" << _projinf->get_user_label(i) << "): ";
    out << armor64A(_projinf->get_user_id(i));
    if (verbose)
      out << " " << _projinf->get_userblk_version(i);
    out << "\n";
  }

  for (unsigned i=0; i<_projinf->nlogs (); i++) {
    if (verbose)
      out << "Log (" << _projinf->get_log_label(i) << "): ";
    out << armor64A (_projinf->get_log_ptr (i));
    if (verbose)
      out << " " << _projinf->get_log_dhash_version(i);
    out << "\n";
  }
  out.tosuio ()->output (1);

  vNew refcounted<traverse>
    (wrap (&fetch_crec_cb), v->commits, wrap (&commit_checker));
  lastv = v->commits;
}

void
start (bigint vid, str server)
{
  _projinf = New refcounted<projinterface>(vid);
  _projinf->fetch_heads(wrap(&got_projinf, vid));
}

static void
usage ()
{
  warnx << "usage: " << progname << " [-v] [-f file] projectid\n";
  exit (1);
}

int
main (int argc, char **argv)
{
  setprogname (argv[0]);
  char *v_file = 0;

  int c;
  while ((c = getopt (argc, argv, "vf:")) != -1)
    switch (c) {
    case 'v':
      verbose = true;
      break;
    case 'f':
      v_file = optarg;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if (argc != 1)
    usage ();

  if (v_file != 0) {
    vfd = open (v_file, O_CREAT | O_RDWR, 0644);
    if (vfd < 0)
      warn << "cannot open " << v_file << ", ignoring it\n";
  }
  if (vfd >= 0) {
    str x = file2wstr (v_file);
    if (x) {
      xin xx (x);
      if (!comvec::parse (&xx, lastv)) {
        warn << "cannot parse a version vector from "
	     << v_file << ", ignoring it\n";
	lastv.clear ();
      }
    }
  }

  bigint vid;
  char sym[1024];
  str dearm;
  rxx path_regexp ("project:([0-9a-zA-Z+-]*)[/]*$");
  int r = readlink(argv[0], sym, 1024);
  if (r > 0) sym[r] = '\0';
  if (path_regexp.search (argv[0]) || (r > 0 && path_regexp.search(sym)))
    dearm = dearmor64A (path_regexp[1]);
  else
    dearm = dearmor64A (argv[0]);
  if (dearm) 
    mpz_set_rawmag_be (&vid, dearm.cstr(), dearm.len());
  else {
    warnx << "pastlslog: key hash invalid\n";
    exit(1);
  }

  nsobj = ns::alloc (vid);
  nsobj->connect (wrap (start, vid));
  amain ();
}

