/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "stdinc.h"
#include "traverse.h"
#include "pack.h"
#include "pastdbfe.h"
#include "build.h"
#include "load.h"
#include "snapshot.h"
#include "replicaimpl.h"

struct br_tree_node;
typedef ptr<br_tree_node> bp;

class scanner : public virtual refcount {
private:
  replica_impl* _replica;
  vec<ptr<snapshot> > prevs;
  cbsnv_t scb;
  bigint projid;
  comvec start;
  vec<ptr<snapshot> > branch_leaves;  // Keeps track of which snapshots we know to be active leaves.
  u_int64_t t_mark;
  //
  // scanning
  //
  vec<crec> commits;
  vec<bigint> commith;
  vec<vec<deltap> > deltas;
  vec<unsigned> ntofetch;
  vec<unsigned> ndeltas;
  log_status loaderr;
  unsigned nloaded;
  unsigned nneeded;
  bool processcrecs;
  bool traversedone;

  vec<bp> bt_roots;
  vec<bp> bt_nodes;

  void scan (const comvec& start);
  void traverse_cb (iterinfo *);
  void load_record_cb (unsigned i, unsigned j, ptr<ns_block> b);
  void load_deltas (crec c, unsigned i);
  void get_deltas_cb (crec c, unsigned i, bigint n, bool failed);
  void load_done ();
  void print_leaves(str s);
  int find_leaf(bigint node);
  void add_leaf(ptr<snapshot> snap);

  //
  // building
  //
  unsigned nbuilds;
  cfsrw_status build_status;
  vec<ptr<snapshot> > sobjs;

  bool add_batch (bp nb);
  void partition ();
  void build_done (cfsrw_status);
  void build_next (bp p);
  void build_one (bp p, ptr<snapshot>, cfsrw_status);
  void load_old_node (bp nb, sptr s, cfsrw_status status);
  bool commit_already_loaded (bigint c);

public:
  scanner (replica_impl *r, bigint projid, const comvec& start, cbsnv_t cb);
  scanner (replica_impl *r, bigint projid, const comvec& start, cbsnv_t cb,
           vec<ptr<snapshot> > p);
};

struct br_tree_node {
  commit_info bv;
  crec cr;
  vec<deltap> deltas;
  ptr<snapshot> base;
  ptr<snapshot> sobj;
  bool alive;
  int alive_children;
  vec<bp> next;
  vec<bool> fork;
  bp prev;

  br_tree_node () {
    cr = 0;
    base = 0;
    sobj = 0;
    alive = true;
    alive_children = 0;
    next.clear ();
    prev = 0;
  }
  ~br_tree_node () {
    prev = 0;
    next.clear ();
    cr = 0;
    base = 0;
    sobj = 0;
  }
};


void
build_snapshot (replica_impl* r, bigint projid, const comvec& start, 
                cbsnv_t cb, vec<sptr> known_branch_heads) 
{
  /* bulid_snapshot calls <cb> with a list of snapshot pointers.
   *   If there are no new entries, or build_snapshot encounters an
   *   error, then the list of snapshot pointers is empty.
   *   If there are new snapshots, then it will return the new snapshots?
   */
  vec<ptr<snapshot> > v;
  while (!known_branch_heads.empty ()) {
    sptr o = known_branch_heads.pop_front ();
    ptr<snapshot> s = obj2impl (o);
    v.push_back (s);
  }
  ptr<scanner> s = New refcounted<scanner> (r, projid, start, cb, v);
}

void
build_snapshot (replica_impl* r, bigint projid, const comvec& start,
                cbsnv_t cb)
{
  ptr<scanner> s = New refcounted<scanner> (r, projid, start, cb);
}

scanner::scanner (replica_impl* r, bigint projid, const comvec& start,
                  cbsnv_t cb)
  : _replica (r), scb (cb), projid (projid), start (start)
{
  loaderr = LOG_OK;
  nloaded = 0;
  nneeded = 0;
  processcrecs = false;
  traversedone = false;
  t_mark = getusec ();
  prevs.clear ();
  scan (this->start);
}

scanner::scanner (replica_impl* r, bigint projid, const comvec& start, 
                  cbsnv_t cb, vec<ptr<snapshot> > known_branch_heads)
  : _replica (r), prevs (known_branch_heads),
    scb (cb), projid (projid), start (start), branch_leaves (known_branch_heads)
{
  loaderr = LOG_OK;
  nloaded = 0;
  nneeded = 0;
  processcrecs = false;
  traversedone = false;
  t_mark = getusec ();
  assert (prevs.size () > 0);
  scan (this->start);
}

bool
scanner::commit_already_loaded (bigint c)
{
  bigint s;
  if (_replica->lookup_snapshot (c, s))
    return true;
  return false;
}

void 
scanner::scan (const comvec& start)
{
  comvec v;
  vNew refcounted<traverse>
    (wrap (mkref (this), &scanner::traverse_cb), start,
           wrap (mkref (this), &scanner::commit_already_loaded));
}

void 
scanner::traverse_cb (iterinfo *li)
{
  if (!li || li->status != LOG_OK) {
    warn << "cannot scan logs to build snapshot (err " << li->status << ")\n";
    scb (vec<sptr> (), CFSRWERR_ERROR);
    return;
  }

  if (!li->finished) {
    crec c = li->commit;
    nneeded++;
    commits.push_back (c);
    commith.push_back (li->hash);
    vec<deltap> v;
    deltas.push_back (v);
    ntofetch.push_back (c->ndeltas ());
    ndeltas.push_back (c->ndeltas ());
    load_deltas (c, commits.size ()-1);
  }

  else {
    traversedone = true;
    load_done ();
  }
}

void
scanner::load_done ()
{
  if (!traversedone)
    return;

  if (nneeded == 0) {
    scb (vec<sptr> (), CFSRWERR_NONEWRECS);
    return;
  }

  else if (nloaded == nneeded && !processcrecs) {
    processcrecs = true;

    if (loaderr) {
      scb (vec<sptr> (), CFSRWERR_IO);
      return;
    }
    if (nloaded > 1)
      warn << "loaded " << nloaded << " commit records\n";

#if 0
    u_int64_t diff = getusec () - t_mark;
    long d = diff/1000;
    warn << "load deltas for snapshot in " << d << " ms\n";
#endif

    // start building the branch tree

    bt_roots.clear ();
    bt_nodes.clear ();

    for (unsigned i=0; i<prevs.size (); i++) {
      bp nb = New refcounted<br_tree_node> ();
      nb->sobj = prevs [i];
      nb->bv = prevs [i]->branch ();
      bt_nodes.push_back (nb);
      bt_roots.push_back (nb);
    }

    partition ();
  }
}

void
scanner::load_record_cb (unsigned i, unsigned j, ptr<ns_block> b)
{
  assert (ntofetch [i] > 0);
  ntofetch [i]--;
  if (!b) {
    loaderr = LOGERR_NORECORD;
    (deltas [i])[j] = 0L;
  }
  else {
    str s (b->data, b->len);
    deltap r = cfsw_deltap_parse (s);
    if (!r) {
      loaderr = LOGERR_NORECORD;
      (deltas [i])[j] = 0L;
    }
    else
      (deltas [i])[j] = r;
  }

  if (ntofetch [i] == 0) {
    assert (ndeltas [i] == 0);
    nloaded++;
    load_done ();
    return;
  }
}

void
scanner::get_deltas_cb (crec c, unsigned i, bigint n, bool failed)
{
  assert (ndeltas [i] > 0);
  assert (ntofetch [i] > 0);
  ndeltas [i]--;

  unsigned j = (deltas [i]).size ();
  (deltas [i]).push_back ((deltap)0L);

  if (failed) {
    ntofetch [i]--;
    loaderr = LOGERR_NORECORD;
    // we won't get any more callbacks from get_deltas
    for (unsigned k=0; k<ndeltas [i]; k++) {
      ntofetch [i]--;
      (deltas [i]).push_back ((deltap)0L);
    }
    ndeltas [i] = 0;
    if (ntofetch [i] == 0) {
      nloaded++;
      load_done ();
      return;
    }
  }
  else
    nsobj->retrieve_block
      (n, wrap (mkref (this), &scanner::load_record_cb, i, j));
}

static void
get_ind_cb (bigint n) {}

void
scanner::load_deltas (crec c, unsigned i)
{
  if (c->ndeltas () > 0)
    get_deltas
      (c, wrap (get_ind_cb),
       wrap (mkref (this), &scanner::get_deltas_cb, c, i));
  else {
    assert (ndeltas [i] == 0);
    assert (ntofetch [i] == 0);
    nloaded++;
    load_done ();
    return;
  }
}

static void
backtrack (bp b)
{
  if (b->prev) {
    bp p = b->prev;
    assert (p->alive_children > 0);
    p->alive_children--;
    if (p->alive_children == 0) {
      p->alive = false;
      backtrack (p);
    }
  }
  else
    b->alive = false;
}

static void
prune (bp b)
{
  if (b->alive == false)
    backtrack (b);
  else {
    for (unsigned i=0; i<b->next.size (); i++)
      prune (b->next [i]);
  }
}

void
scanner::load_old_node (bp nb, sptr s, cfsrw_status status)
{
  if (!s || status) {
    commit_info inf = nb->bv;
    warn << "Cannot find dependence for commit record(1) "
         << armor64A (inf.rptr) << "\n";
    inf = nb->cr->depends();
    warn << "Claim to depend on commit record "
         << armor64A (inf.rptr) << "\n";
    scb (vec<sptr> (), CFSRWERR_ERROR);
    return;
  }

  bp b = New refcounted<br_tree_node> ();
  b->sobj = obj2impl (s);
  b->bv = s->branch ();
  bt_nodes.push_back (b);
  bt_roots.push_back (b);
  prevs.push_back (obj2impl (s));

  if (!add_batch (nb)) {
    crec cr = nb->cr;
    commit_info inf = nb->bv;
    warn << "Cannot find dependence for commit record(2) "
         << armor64A (inf.rptr) << "\n";
    inf = cr->depends();
    warn << "Claim to depend on commit record "
         << armor64A (inf.rptr) << " n";
    scb (vec<sptr> (), CFSRWERR_ERROR);
    return;
  }

  partition ();
}

bool
scanner::add_batch (bp nb)
{
  crec cr = nb->cr;

  // look for dependencies
  for (unsigned i=0; i < bt_nodes.size (); i++) {
    bp b = bt_nodes [i];
    
    bigint x1 = cr->depends ().rptr;
    bigint x2 = b->bv.rptr;

    if (cr->depends ().rptr == b->bv.rptr) { // depends on "b"
      b->alive = true;
      b->alive_children++;
      nb->prev = b;
      b->next.push_back (nb);

      if (cr->flags () & commit_record::BRANCH_OFF) // new branch
	b->fork.push_back (true);
      else
	b->fork.push_back (false);
      bt_nodes.push_back (nb);
      return true;
    }
  }

  if (prevs.size () == 0) {
    //warn << "root of branch tree " << armor64A (nb->bv.rptr) << "\n";
    // new root node
    nb->base = 0L;
    nb->prev = 0L;
    bt_nodes.push_back (nb);
    bt_roots.push_back (nb);
    return true;
  }

  return false;
}

void
scanner::partition ()
{
  if (commits.size () > 0) {
    bp nb = New refcounted<br_tree_node> ();
    nb->cr = commits.pop_front ();
    nb->bv.head = nb->cr->head ();
    nb->bv.rptr = commith.pop_front ();
    nb->deltas = deltas.pop_front ();
    nb->base = 0;
    nb->sobj = 0;

    if (!add_batch (nb)) {
      //warn << "new batch of deltas from a different partition\n";
      bigint sn_addr = _replica->crec_to_snapaddr(nb->cr->depends().rptr);
      load_snapshot (sn_addr, wrap (mkref (this), &scanner::load_old_node, nb));
      
      // This means that it depends on a commit record that is
      // not currently a leaf, or is not a new commit record.
      // We have to load up an older snapshot for this.

      // Actually, we don't want to load the snapshot, we want to load the commit record.

      return;
    }

    // recurse until we have no more batches...
    partition ();
  }

  else { // no more batches...
    // prune the tree to get rid of dead branches
    for (unsigned i=0; i<bt_roots.size (); i++) {
      bp b = bt_roots [i];
      prune (b);
    }

    nbuilds = 0;
    nbuilds++;
    build_status = CFSRW_OK;
    for (unsigned i=0; i<bt_roots.size (); i++) {
      if (bt_roots [i]->alive) {
	build_next (bt_roots [i]);
      }
    }
    nbuilds--;
    build_done (CFSRW_OK);
  }
}

static bp
find_branch_pt (bp b)
{
  assert (b->next.size () > 0);
  if (b->next.size () > 1)
    return b;
  return find_branch_pt (b->next [0]);
}

void
scanner::build_done (cfsrw_status status)
{
  if (status)
    build_status = status;

  if (nbuilds == 0) {
    vec<sptr> v;
    while (!branch_leaves.empty ()) {
      ptr<snapshot> s = branch_leaves.pop_front ();
      v.push_back (s);
    }
    scb (v, build_status);
  }
}

void
scanner::build_next (bp p)
{
  nbuilds++;
  if (p->sobj == 0)
    vNew refcounted<bob>
      (_replica, projid, p->base, p->deltas, p->cr, p->bv,
       wrap (mkref (this), &scanner::build_one, p));
  else
    build_one (p, p->sobj, CFSRW_OK);
}

void
scanner::build_one (bp p, ptr<snapshot> nobj, cfsrw_status status)
{
  assert (nbuilds > 0);
  commit_info x = nobj->branch();
  bool is_new_snap = !commit_already_loaded(x.rptr);
  nbuilds--;
  p->sobj = nobj;
  if (p->base && nobj &&
      (!p->cr || (p->cr->flags () & commit_record::BRANCH_OFF) == 0))
    p->sobj->ptag = p->base->ptag;

  if (!nobj || status) {
    warn << "build failed, err " << status << "\n";
    build_done (status);
    return;
  }

  if (p->cr) {
    p->bv.vptr = nobj->addr ();
    btree_node n;
    n.commit = p->bv;
    n.level = 0;
    n.proposed_name = p->cr->proposed_name ();
    n.prev = p->cr->depends ();
    _replica->btree ()->add_node (p->bv.rptr, n);
  }

  if (is_new_snap) {
    // only mutate the leafset if the snapshot is new.

    // Figure out how this new snapshot changes the existing
    // list of branch leaves.

    // If this is an explicit branch, add it to the list
    // else if parent is a leaf, then this replaces it
    // else if parent is not leaf, then just add this

    int x;

    if (!nobj->prev()) {
      // This is the root snapshot
      add_leaf(p->sobj);
      
    } else if (p->cr && p->cr->flags() & commit_record::BRANCH_OFF) {
      add_leaf(p->sobj);
      
    } else if ( (x = find_leaf(p->base->addr())) >= 0) {
      branch_leaves[x] = p->sobj;
      
    } else if (find_leaf(p->base->addr()) < 0) {
      // Base snapshot is not a leaf: this snapshot is an implicit branch
      add_leaf(p->sobj);
    
    } else {
      warn << "  This should not occur\n";
      assert(0);
    }    
  }

  if (p->next.size () > 0) {
    for (unsigned i=0; i<p->next.size (); i++) {
      if (p->next [i]->alive) {
        p->next [i]->base = p->sobj;
        build_next (p->next [i]);
      }
    }

    // remove reference, so we don't keep too many inmem dbs open
    p->sobj = 0L;
  }
  build_done (CFSRW_OK);
}
/*
void
scanner::print_leaves(str s) {
  int i;
  for(i=0; i<(int)branch_leaves.size(); i++)
    warn << s << ": " << i << " " << armor64A(branch_leaves[i]->addr()) << "\n";
}
*/
int
scanner::find_leaf(bigint node) {
  int i;
  for(i=0; i<(int)branch_leaves.size(); i++)
    if (branch_leaves[i]->addr() == node)
      return i;
  return -1;
}

void
scanner::add_leaf(ptr<snapshot> snap) {
  if (find_leaf(snap->addr()) < 0) {
    branch_leaves.push_back(snap);
  }
}
