
#include <sfscrypt.h>
#include <chord.h>
#include <id_utils.h>
#include <dbfe.h>
#include <merkle_misc.h>

#include "verify.h"
#include "keyhash_idx.h"

static inline bigint
get_projkey (ptr<dbrec> d)
{
  bigint n;
  int r = keyhash_projkey (d, n);
  assert (r == 0);
  return n;
}

keyhash_idx::keyhash_idx (ptr<dhash_blockstore> storage)
  : _syncnext (0), _vidx (400000), _storage (storage)
{
  unsigned int i;
  vec < ptr<dbrec> > key_list;
  assert(!_storage->getpkblockids(&key_list));

  for(i=0; i<key_list.size(); i++) {
    bigint n = dbrec2id (key_list[i]);
    ptr<dbrec> d = _storage->lookup_pkblock(key_list[i]);
    assert (d);
    bigint lk = get_projkey (d);
    long v = keyhash_version (d);
    _vidx.insert (n, v);
    lidx_add (lk, n);
  }
}

void
keyhash_idx::lidx_add (bigint projkey, bigint key)
{
  lwrap *l = _lidx.search (projkey);
  if (l) {
    // if we offered this set of keys to someone else, mark the keys
    // as not offered because we just got a new key!
    l->offered = false;
    if (l->t->search (key) == NULL) {
      khwrap *k = New khwrap (key);
      l->t->insert (k);
    }
  }
  else {
    l = New lwrap (projkey);
    khwrap *k = New khwrap (key);
    l->t->insert (k);
    _lidx.insert (l);
  }
}

int
keyhash_idx::lidx_remove (bigint projkey, bigint key)
{
  lwrap *l = _lidx.search (projkey);
  if (l) {
    if (l->t->remove(key)) {
      if (l->t->size() == 0) {

	if (_syncnext && _syncnext->n == l->n) {
	  if (_syncnext == _lidx.last ())
	    _syncnext = _lidx.first ();
	  else
	    _syncnext = _lidx.next (_syncnext);
	}
	if (_syncnext && _syncnext->n == l->n)
	  _syncnext = NULL;
	
	_lidx.remove(projkey);
	delete (l);
      }
      return 0;
    }    
  } else {
    warn << "keyhash_idx::lidx_remove Could not find project: "
	 << armor64A(projkey) << "\n";
  }
  return -1;
}

dhash_stat
keyhash_idx::insert (ptr<dbrec> key, ptr<dbrec> val)
{
  return insert (dbrec2id (key), key, val);
}

dhash_stat
keyhash_idx::insert (bigint projkey, ptr<dbrec> key, ptr<dbrec> val)
{
  bigint n = dbrec2id (key);
  long v = keyhash_version (val);
  bigint k;
  int r = keyhash_projkey (val, k);
  if (r != 0 || k != projkey) {
    if (r != 0)
      warn << "can't get projkey from block\n";
    else
      warn << "block has proj key " << k
	   << ", but insert claims it is " << projkey << "\n";
    return DHASH_STORE_NOVERIFY;
  }

  ptr<dbrec> prev = _storage->lookup_pkblock (key);
  if (prev) {
    long vp = keyhash_version (prev);
    if (vp >= v)
      return DHASH_STALE;
    else
      _storage->remove_pkblock(key);
  }

  r = _storage->insert_pkblock (projkey, key, val);
  if (r)
    return DHASH_STOREERR;

  _vidx.remove (n);
  _vidx.insert (n, v);
  lidx_add (projkey, n);

  return DHASH_OK;
}

// XXX lookup should insert the retrieved block into the 
//     indexes because it may have been inserted directly into
//     the database. This must be added when merkle syncing is
//     separated from lsd.

ptr<dbrec>
keyhash_idx::lookup (bigint n)
{
  ptr<dbrec> d = _storage->lookup_pkblock (id2dbrec (n));
  if (!d)
    return 0L;
  return d;
}

ptr<dbrec>
keyhash_idx::lookup (bigint n, bigint &lk)
{
  ptr<dbrec> d = _storage->lookup_pkblock (id2dbrec (n));
  if (!d)
    return 0L;
  lk = get_projkey (d);
  return d;
}

bool
keyhash_idx::getprojkey (bigint n, bigint &l)
{
  ptr<dbrec> d = _storage->lookup_pkblock (id2dbrec (n));
  if (!d)
    return false;
  l = get_projkey (d);
  return true;
}

bool
keyhash_idx::getver (bigint n, long &v)
{
  long *i = _vidx [n];
  if (i) {
    v = *i;
    return true;
  }

  ptr<dbrec> d = _storage->lookup_pkblock (id2dbrec (n));
  if (!d)
    return false;

  warn << "poking db for version data\n";
  v = keyhash_version (d);
  _vidx.insert (n, v);
  return true;
}

static void
findkeys_traverse_k (keyhash_idx *o, ptr<keyhash_idx::offer_state> s,
                     khwrap *k)
{
  bigint n = k->n;
  long v;
  bool r = o->getver (n, v);
  assert (r);
  s->keys.push_back (n);
  s->vers.push_back (v);
}

ptr<keyhash_idx::offer_state>
keyhash_idx::findkeys (bigint pred, bigint my)
{
  // find all keys stored under keys (pred, my]

  ptr<offer_state> s = New refcounted<offer_state> ();
  lwrap *l, *first;
  first = l = _lidx.closestsucc (pred);
  bool sawfirst = false;

  while (l && 
	 betweenrightincl(pred, my, l->n) && 
	 (l != first || !sawfirst) ) {
    sawfirst = true;
    l->t->traverse (wrap (findkeys_traverse_k, this, s));
    l = _lidx.closestsucc (l->n + 1);
  }
  return s;
}

ptr<keyhash_idx::offer_state>
keyhash_idx::offernext (bigint pred, bigint my)
{
  if (_syncnext == 0L || _syncnext == _lidx.last ())
    _syncnext = _lidx.first ();
  else
    _syncnext = _lidx.next (_syncnext);

  if (!_syncnext)
    return 0;

  bigint start = _syncnext->n;
  struct timeval tv;
  gettimeofday (&tv, NULL);

  do {
    if (!between (pred, my, _syncnext->n) &&
	(!_syncnext->offered ||
	 (_syncnext->offered && !_syncnext->notified &&
	  ((tv.tv_sec - _syncnext->offertime) > 5*60)))) {
      ptr<offer_state> s = New refcounted<offer_state> ();
      _syncnext->t->traverse (wrap (findkeys_traverse_k, this, s));
      //_syncnext->offered = true;
      _syncnext->offertime = tv.tv_sec;
      s->projkey = _syncnext->n;
      return s;
    }
    if (_syncnext == _lidx.last ())
      _syncnext = _lidx.first ();
    else
      _syncnext = _lidx.next (_syncnext);
  }
  while (_syncnext->n != start);

  return 0;
}

void
keyhash_idx::notify (bigint k, bool set)
{
  lwrap *l = _lidx.search (k);
  if (l) {
    if (set && l->offered)
      l->notified = true;
    else {
      l->offered = false;
      l->notified = false;
    }
  }
}

int
keyhash_idx::remove(bigint removeme) {
  bigint projkey;
  if (!getprojkey(removeme, projkey)) {
    warn << "keyhash_idx::remove could not find project key for block: " 
	 << armor64A(removeme) << "\n";
    return -1;
  }

  lidx_remove(projkey, removeme);
  _vidx.remove(removeme);

  return 0;
}
