/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "arpc.h"
#include "sfsconnect.h"
#include "sfsclient.h"
#include "stdinc.h"
#include "penviron.h"

static int conn_timer = atoi(pgetenv("PASTWATCH_CONNTIMER"));
static int debug_conn = atoi(pgetenv("PASTWATCH_DEBUG_CONN"));

struct multiconn_args
{
  vec<str> hosts;
  vec<int> ports;
  vec<int> responses;
  int timeout;
  bool timedout;
  callback<void,int,str>::ptr cb;
  str connected;
  timecb_t* timeout_timer;

  multiconn_args (vec<str> h, vec<int> p, int t, callback<void,int,str>::ptr c)
  {
    unsigned int i;
    hosts = h;
    ports = p;
    timeout = t;
    timedout = false;
    cb = c;
    connected = "";
    timeout_timer = NULL;
    for(i=0; i<h.size(); i++) 
      responses.push_back(0);
  }

  void saw_response (int i) {
    responses[i] = 1;
  }

  bool all_responded () {
    for(unsigned int i=0; i<responses.size(); i++) {
      if (responses[i] == 0) return false;
    } 
    return true;
  }
};

void
multiconnect_cb (ptr<multiconn_args> args, unsigned int n, int fd)
{
  if (args->timedout) {
    if (fd >= 0)
      close (fd);
    return;
  }

  args->saw_response (n);

  if (fd < 0) {
    if (args->connected == "") {
      warn << "connect to: " << args->hosts[n]
           << ":" << args->ports[n] << " failed.\n";
      if (args->all_responded ())
        (args->cb) (-1, "");
    }
    return;
  }

  else if (args->connected != "") {
    // redundant connection established. Kill it.
    close (fd);
    return;
  }

  else {
    if (debug_conn) 
      warn << "Connected to: " << args->hosts[n] << ":" << args->ports[n] << "\n";
    timecb_remove (args->timeout_timer);
    args->connected = strbuf() <<  args->hosts[n] << ":" << args->ports[n];
    (args->cb) (fd, args->hosts[n]);
  }
}

void
multiconnect_timeout (ptr<multiconn_args> args)
{
  warn << "Connect to Aqua nodes: Timed out.  To increase "
       << "timeout period, see PASTWATCH_CONNTIMER in the pastwatch(1) "
       << "man page\n"; 
  args->timedout = true;
  if (args->connected == "" && !args->all_responded ())
    (args->cb) (-1, "");
}

void 
multiconnect_real (ptr<multiconn_args> args)
{
  if (debug_conn)
    warn << "Trying to connect to an Aqua node\n";
  args->timeout_timer = delaycb (args->timeout, 0, wrap (&multiconnect_timeout, args));

  for (unsigned i=0; i<args->hosts.size (); i++)
    tcpconnect
      (args->hosts[i], args->ports[i], wrap (&multiconnect_cb, args, i));
}

void
multiconnect (vec<str> hosts, vec<int> ports, int timeout, callback<void,int,str>::ptr cb)
{
  // tries hosts/ports in parallel. calls <cb> once the first
  // connection is established, and closes all other successful
  // connections.  fail if connections cannot be established within
  // <timeout> seconds.
  assert(hosts.size() == ports.size());
  assert(hosts.size() > 0);
  ptr<multiconn_args> args =
    New refcounted<multiconn_args> (hosts, ports, timeout, cb);
  multiconnect_real (args);
}

static void
parse_add_host (const char* s, vec<str>& hv, vec<int>& pv)
{
  char *p = strchr (s, ':');
  if (!p) {
    warn << "parse add host ." << s << ".\n";
    warn << "no port, skip\n";
    return;
  }

  char x [p-s+1];
  strncpy (x, s, p-s);
  x [p-s] = '\0';

  p++;
  if (inet_addr (x) == INADDR_NONE) {
    struct hostent *h;
    h = gethostbyname (x);
    if (!h) {
      warn << "Invalid address or hostname: " << x << "\n";
      exit (1);
    }
    struct in_addr *ptr = (struct in_addr *) h->h_addr;
    hv.push_back (inet_ntoa (*ptr));
  }
  else
    hv.push_back (x);
  
  pv.push_back (atoi (p));
}

void
proxy_connected (callback<void,str,ptr<axprt_stream> >::ptr cb, int fd, str server)
{
  if (fd < 0)
    cb ("", (ptr<axprt_stream>) 0L);
  else {
    ref<axprt_stream> xx = axprt_stream::alloc (fd, 1024*1025);
    cb (server, xx);
  }
}

void
make_connection (const char *arg, callback<void,str,ptr<axprt_stream> >::ref cb)
{
  vec<str> hosts;
  vec<int> ports;

  char x [strlen (arg)+1];
  strncpy (x, arg, strlen (arg)+1);

  char *h = x;
  char *c = 0L;

  while((c = strchr (h, ','))) {
    parse_add_host (h, hosts, ports);
    h = c+1;
  }
  if (strlen (h))
    parse_add_host (h, hosts, ports);

  if (hosts.size() > 0)
    multiconnect
      (hosts, ports, conn_timer, wrap (proxy_connected, cb));
  else
    cb ("", (ptr<axprt_stream>) 0L);
}

