/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef VERVEC_H
#define VERVEC_H

#include "stdinc.h"
#include "pack.h"

class vervec {
protected:
  vec<bigint> _head;
  vec<uint> _seqn;

  void add_no_check (bigint h, uint s) {
    _head.push_back(h);
    _seqn.push_back(s);
  }

public:
  vervec() { clear(); }
  ~vervec () { clear (); }
  void clear () { _head.clear(); _seqn.clear(); }

  bool empty () const { return _head.size() == 0; }
  unsigned size () const { return _head.size(); }
  bigint head (int i) const { return _head[i]; }
  uint seqn (int i) const { return _seqn[i]; }

  void print () const {
    warn << "vervec:\n";
    for(unsigned i=0; i<_head.size(); i++)
      warn << "  " << armor64A(_head[i]) << " : " << _seqn[i] << "\n";
  }

  void serialize (strbuf &sb) const {
    sint(sb, _head.size());
    for (unsigned i=0; i<_head.size(); i++) {
      sbigint (sb, _head[i]);
      sint (sb, _seqn[i]);
    }
  }

  vervec &operator= (const vervec &v) {
    clear();
    for (unsigned i=0; i<v._head.size(); i++) {
      _head.push_back(v._head[i]);
      _seqn.push_back(v._seqn[i]);
    }
    return *this;
  }

  void clear (bigint h)
  {
    vervec v;
    for (unsigned i=0; i<_head.size (); i++)
      if (_head [i] != h)
	v.add_no_check (_head [i], _seqn [i]);
    *this = v;
  }

  // v0 equals v1 iff (forall i, v0[i] == v1[i])
  bool equals (const vervec& v) const {
    unsigned i;
    unsigned j;
    for (i=0, j=0; i<_head.size();) {
      if (j < v._head.size()) {
        if (_head[i] == v._head[j]) { // both vector contain same user
	  if (_seqn[i] != v._seqn[j])
	    return false;
	  j++;
	  i++;
        }
        else if (_head[i] < v._head[j]) // user in mine, but not hers
	  return false;
        else // in hers, but not in mine
	  return false;
      }
      else // in mine, not hers
	return false;
    }
    if (j < v._head.size()) // rest of users in hers, but not mine
      return false;
    return true;
  }
  
  void add (bigint h, uint s)
  {
    // make sure heads are in the same order as they are in the projinf:
    // which is from smallest to largest as defined by the > operator
    // on bigint
    bool inorder = true;
    unsigned i = 0;
    for (i=0; i<_head.size(); i++) {
      if (_head [i] == h) {
	_seqn [i] = s;
	return;
      }
      else if (_head[i] > h) {
        inorder = false;
	break;
      }
    }
    if (inorder) {
      _head.push_back (h);
      _seqn.push_back (s);
    }
    else {
      vec<bigint> nh;
      vec<uint> ns;
      bool added = false;
      for(unsigned j=0; j<_head.size(); j++) {
	if (!added && _head[j] > h) {
          added = true;
	  nh.push_back (h);
	  ns.push_back (s);
	}
	nh.push_back (_head [j]);
	ns.push_back (_seqn [j]);
      }
      _head = nh;
      _seqn = ns;
    }
  }

  static bool parse (xin *in, vervec &version)
  {
    int vsize;
    version.clear ();
    if (!in->rint (vsize))
      return false;
    for (int i=0; i<vsize; i++) {
      bigint h;
      int s;
      if (!in->rbigint (h))
	return false;
      if (!in->rint (s))
	return false;
      version.add_no_check (h, s);
    }
    return true;
  }
};

typedef ptr<vervec> vervecp;

class comvec {
protected:
  vec<bigint> _head;
  vec<bigint> _rptr;

  void add_no_check (bigint h, bigint l) {
    if (_head.size ()) { assert (h > _head [_head.size ()-1]); }
    _head.push_back (h);
    _rptr.push_back (l);
  }

public:
  comvec() { clear(); }
  ~comvec () { clear (); }
  void clear () { _head.clear (); _rptr.clear (); }

  bool empty () const { return _head.size() == 0; }
  unsigned size () const { return _head.size(); }
  bigint head (int i) const { return _head[i]; }
  bigint rptr (int i) const { return _rptr[i]; }

  comvec filter (const comvec& v) const {
    comvec f;
    for (unsigned i = 0; i < v.size (); i++) {
      for (unsigned j = 0; j < _head.size (); j++) {
        if (_head [j] == v.head (i)) {
	  f.add (_head [j], _rptr [j]);
	  break;
	}
      }
    }
    return f;
  }

  // return a version vector that contains entries in *this that are
  // also in v, and entries in v not in *this but with hash set to 0.
  comvec filter_expand (const comvec &v) const {
    comvec f;
    for (unsigned i = 0; i < v.size (); i++) {
      bool found = false;
      for (unsigned j = 0; j < _head.size (); j++) {
        if (_head [j] == v.head (i)) {
	  found = true;
	  f.add (_head [j], _rptr [j]);
	  break;
	}
      }
      if (!found)
	f.add (v.head (i), 0);
    }
    return f;
  }

  void print() const {
    warn << "Comvec:\n";
    for(unsigned i=0; i<_head.size(); i++)
      warn << "  " << armor64A(_head[i]) << " : " << armor64A(_rptr[i]) << "\n";
  }

  void serialize (strbuf &sb) const {
    sint(sb, _head.size());
    for (unsigned i=0; i<_head.size(); i++) {
      sbigint (sb, _head[i]);
      sbigint (sb, _rptr[i]);
    }
  }

  comvec &operator= (const comvec &v) {
    clear();
    for (unsigned i=0; i<v._head.size(); i++) {
      _head.push_back (v._head[i]);
      _rptr.push_back (v._rptr[i]);
    }
    return *this;
  }

  void clear (bigint h)
  {
    comvec v;
    for (unsigned i=0; i<_head.size (); i++)
      if (_head [i] != h)
	v.add (_head [i], _rptr [i]);
    *this = v;
  }

  void add (bigint h, bigint l)
  {
    // make sure heads are in the same order as they are in the projinf:
    // which is from smallest to largest as defined by the > operator
    // on bigint
    bool inorder = true;
    unsigned i = 0;
    for (i=0; i<_head.size(); i++) {
      if (_head[i] == h) {
	_rptr[i] = l;
	return;
      }
      else if (_head[i] > h) {
        inorder = false;
	break;
      }
    }
    if (inorder) {
      _head.push_back(h);
      _rptr.push_back(l);
    }
    else {
      vec<bigint> nh;
      vec<bigint> nl;
      bool added = false;
      for(unsigned j=0; j<_head.size(); j++) {
	if (!added && _head[j] > h) {
          added = true;
	  nh.push_back(h);
	  nl.push_back(l);
	}
	nh.push_back(_head[j]);
	nl.push_back(_rptr[j]);
      }
      _head = nh;
      _rptr = nl;
    }
  }
  
  static bool parse (xin *in, comvec &version)
  {
    int vsize;
    version.clear ();
    if (!in->rint (vsize))
      return false;
    for (int i=0; i<vsize; i++) {
      bigint h;
      bigint l;
      if (!in->rbigint (h))
	return false;
      if (!in->rbigint (l))
	return false;
      version.add_no_check (h, l);
    }
    return true;
  }
};

typedef ptr<comvec> comvecp;

#endif

