/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>
#include "pastwatch.h"
#include "update.h"
#include "path.h"
#include "pdb.h"
#include "io.h"

extern ptr<cfsrw> _cfsrw;

extern void update (bigint root, str path, cbi cb);
extern void checkin (const pde_t& e, callback<void, bigint, int>::ref fhcb);
extern void relink (const pde_t& e, callback<void, bigint, int>::ref fhcb);

extern void
read_file (int, bigint, bigint, callback<void, cfsrw_status>::ref);
extern mtime_t
write_file (str, bigint, bigint, callback<void, cfsrw_status>::ref);
extern int
diff_normal (const char *f1, const char *f2, strbuf &b);
extern mtime_t get_mtime (str);

extern vec<str> notify_cmds;
extern vec<str> notify_files;

static bool
is_binary (str f)
{
  int fd = open (f.cstr (), O_RDONLY);
  if (fd) {
    char buf [512];
    unsigned n = read (fd, buf, 512);
    for (unsigned i=0; i<n; i++) {
      if (buf [i] & 0x80)
	return true;
    }
    return false;
  }
  return true;
}


class commit_object : public virtual refcount
{
  bigint va;
  bigint root;
  int work;
  int error;
  ptr<update_object_spec> up;
  vec<str> committed;
  vec<char> committed_type;
  int stage;

  void done (cbi cb, int error)
  {
    if (!error) {
      /*
       * stages: 
       *   0 = remove
       *   1 = add
       *   2 = commit
       */
      stage++;
      if (stage == 1) {
        work++;
        do_add (cb);
        work--;
        if (work == 0)
	  done (cb, error);
      }
      else if (stage == 2) {
        work++;
        do_commit (cb);
        work--;
        if (work == 0)
	  done (cb, error);
      }
      else if (stage > 2) {
	// if there are errors, driver will abort all our changes, so
	// don't remove any scheduled dirops
        if (!error)
          del_dirops (committed, committed_type);
        cb (error);
      }
      else
        assert (0);
    }
    else
      cb (error);
  }

  void
  cfsrw_reply (cbi cb, cfswcallp c, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "commit: cannot write file to repository\n";
      error = -1;
    }
    if (work == 0)
      done (cb, error);
  }

  void
  write_reply (cbi cb, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "commit: cannot write file to repository\n";
      error = -1;
    }
    if (work == 0)
      done (cb, error);
  }

  void
  symlink_reply (unsigned int idx, cbi cb, cfswcallp c, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "commit: cannot create symlink in repository\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }

    cfsw_symlink_res *res = static_cast<cfsw_symlink_res *> (c->res ());
    pde_t *de = &up->to_commit [idx];
    pmgr->schedule_add (de->path, pfe (res->object));
    pmgr->schedule_add (get_parent_dir (de->path), pfe (de->dirfh));

    if (work == 0)
      done (cb, error);
  }
  
  void
  sa_sattr_reply (unsigned int idx, cbi cb, cfswcallp c, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "commit: cannot set file attribute in repository for setattr\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }

    pde_t *de = &up->to_setattr [idx];
    pastwatch_status ("Setting attributes of", de->path);
    // for commit notification message
    notify_cmds.push_back ("m");
    notify_files.push_back (de->cfspath);
    pmgr->schedule_add (de->path, pfe (de->fh));
    
    if (work == 0)
      done (cb, error);
  }

  void
  sattr_reply (unsigned int idx, cbi cb, cfswcallp c, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "commit: cannot set file attribute in repository for commit\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }

    pde_t *de = &up->to_commit [idx];
    work++;
    mtime_t m = write_file
      (de->path, de->fh, va,
       wrap (mkref(this), &commit_object::write_reply, cb));

    pastwatch_status ("Committing", de->path);
    // for commit notification message
    notify_cmds.push_back ("M");
    notify_files.push_back (de->cfspath);
    pmgr->schedule_add (de->path, pfe (de->fh, m));
  }

  void
  read_file_cb (unsigned int idx, cbi cb, str tfn, int fd, cfsrw_status status)
  {
    work--;
    close (fd);

    if (status) {
      warn << "commit: cannot get previous version of file from repository\n";
      unlink (tfn);
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }

    pde_t *de = &up->to_commit [idx];

    if (is_binary (tfn) || is_binary (de->path)) {
      // binary file, do whole file write
      do_write (idx, cb);
      return;
    }

    mtime_t m = get_mtime (de->path);
    strbuf b;
    int r = diff_normal (tfn, de->path, b);
    unlink (tfn);

    if (r < 0) {
      warn << "commit: cannot generate patch, writing whole file instead\n";
      do_write (idx, cb);
      return;
    }

    work++;
    cfswcallp c = New refcounted<cfswcall>
      (CFSWPROC_PATCH,
       wrap (mkref (this), &commit_object::cfsrw_reply, cb));
    cfsw_patch_arg *arg = static_cast<cfsw_patch_arg *> (c->arg ());
    arg->file = de->fh;
    arg->patch = b;
    cfsw_call (c);

    pastwatch_status ("Committing", de->path);
    // for commit notification message
    notify_cmds.push_back ("M");
    notify_files.push_back (de->cfspath);
    pmgr->schedule_add (de->path, pfe (de->fh, m));
  }

  void
  unlink_reply (unsigned idx, cbi cb, cfswcallp cc, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "commit: cannot unlink symlink in repository for commit\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }
  
    pde_t *de = &up->to_commit [idx];
  
    char buf[256];
    int r;
    if ((r = readlink (de->path, buf, 256)) < 0) {
      perror ("readlink");
      error = -1;
    }
    else {
      str link (buf, r);
      cfswcallp c = New refcounted<cfswcall>
        (CFSWPROC_SYMLINK,
	 wrap (mkref(this), &commit_object::symlink_reply, idx, cb));
      cfsw_symlink_arg *arg = static_cast<cfsw_symlink_arg *> (c->arg ());
      arg->dir = de->dirfh;
      arg->name = de->name;
      arg->linkto = link;
      arg->attributes.valid_size = 1;
      arg->attributes.size = link.len ();
  
      work++;
      cfsw_call (c);

      pastwatch_status ("Committing", de->path);
      // for commit notification message
      notify_cmds.push_back ("M");
      notify_files.push_back (de->cfspath);
    }
  
    if (work == 0)
      done (cb, error);
  }

  void do_commit (cbi cb)
  {
    for (unsigned i=0; i<up->to_commit.size (); i++) {
      pde_t *de = &up->to_commit [i];
     
      if (de->type == CFSLNK) {
	cfswcallp c = New refcounted<cfswcall>
	  (CFSWPROC_UNLINK,
	   wrap (mkref(this), &commit_object::unlink_reply, i, cb));
	cfsw_unlink_arg *arg = static_cast<cfsw_unlink_arg *> (c->arg ());
	arg->object = de->fh;
	if (de->type == CFSLNK)
	  arg->type = CFSW_LNK;
	else if (de->type == CFSDIR)
	  arg->type = CFSW_DIR;
	else if (de->type == CFSREG)
	  arg->type = CFSW_REG;
	else
	  assert (0);
        arg->dir = de->dirfh;
        arg->name = de->name;
          
	work++;
        cfsw_call (c);
      }
      else if (de->type == CFSREG) {
	str s = pastwatch_tmpfile ();
        int fd = open (s, O_CREAT | O_TRUNC | O_WRONLY, 0600);
        bigint cr_addr = my_branch_head (_cfsrw->getreplica ());
	work++;
        read_file
          (fd, de->fh, cr_addr,
           wrap (mkref (this), &commit_object::read_file_cb, i, cb, s, fd));
      }
      else
	assert (0);
    }
    for (unsigned i=0; i<up->to_setattr.size (); i++) {
      pde_t *de = &up->to_setattr [i];

      cfswcallp c = New refcounted<cfswcall>
        (CFSWPROC_SETATTR,
         wrap (mkref(this), &commit_object::sa_sattr_reply, i, cb));
      cfsw_setattr_arg *arg =
        static_cast<cfsw_setattr_arg *> (c->arg ());

      arg->object = de->fh;
      arg->new_attributes.valid_mode = 1;
      arg->new_attributes.mode = de->mode;
      work++;
      cfsw_call (c);
    }
  }

  void
  do_write (unsigned i, cbi cb)
  {
    pde_t *de = &up->to_commit [i];
    cfswcallp c = New refcounted<cfswcall>
      (CFSWPROC_SETATTR,
       wrap (mkref(this), &commit_object::sattr_reply, i, cb));
    cfsw_setattr_arg *arg =
      static_cast<cfsw_setattr_arg *> (c->arg ());

    arg->object = de->fh;
    arg->new_attributes.valid_size = 1;
    arg->new_attributes.size = 0;
    arg->new_attributes.valid_mode = 1;
    arg->new_attributes.mode = de->mode;
    work++;
    cfsw_call (c);
  }

  void
  remove_unlink_reply (int i, cbi cb, cfswcallp cc, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "commit: cannot lookup "
	   << up->to_remove [i].path << " in repository\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }
    pastwatch_status ("Removing", up->to_remove [i].path);
    // for commit notification message
    notify_cmds.push_back ("R");
    notify_files.push_back (up->to_remove [i].cfspath);

    committed.push_back (up->to_remove [i].path);
    committed_type.push_back ('R');
    pmgr->schedule_remove (up->to_remove [i].path);

    str path = up->to_remove [i].path;
    str parent = get_parent_dir (path);
    ptr<pfe> x = pmgr->lookup (parent);

    if (x) {
      x->cr_addr = 0;
      pmgr->schedule_add (parent, *x);
    } 

    if (work == 0)
      done (cb, error);
  }

  void remove_lookup_reply (int i, cbi cb, cfsrcallp cc, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << "commit: cannot lookup "
	   << up->to_remove [i].path << " in repository\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }

    cfsr_lookup_arg *arg = static_cast<cfsr_lookup_arg *> (cc->arg ());
    cfsr_lookup_res *res = static_cast<cfsr_lookup_res *> (cc->res ());

    cfswcallp c = New refcounted<cfswcall>
      (CFSWPROC_UNLINK,
       wrap (mkref(this), &commit_object::remove_unlink_reply, i, cb));
    cfsw_unlink_arg *a = static_cast<cfsw_unlink_arg *> (c->arg ());
    a->object = res->object;
    if (up->to_remove [i].type == CFSLNK)
      a->type = CFSW_LNK;
    else if (up->to_remove [i].type == CFSDIR)
      a->type = CFSW_DIR;
    else if (up->to_remove [i].type == CFSREG)
      a->type = CFSW_REG;
    else
      assert (0);
    a->dir = arg->dir;
    a->name = arg->name;
    work++;
    cfsw_call (c);
  }

  void do_remove (cbi cb)
  {
    for (unsigned i=0; i<up->to_remove.size (); i++) {
      work++;
      cfsr_lookup
	(root, up->to_remove [i].cfspath,
	 wrap (mkref(this), &commit_object::remove_lookup_reply, i, cb));
    }
  }
 
  void 
  relink_reply (int i, cbi cb, bigint fh, int status)
  {
    work--;
    if (status) {
      warn << "commit: cannot create "
	   << up->to_add [i].path << " in repository\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }
    
    pastwatch_status ("Relinking", up->to_add [i].path);
    // for commit notification message
    notify_cmds.push_back ("A");
    notify_files.push_back (up->to_add [i].path);

    pde_t *de = &up->to_add [i];
    committed.push_back (de->path);
    committed_type.push_back ('A');
    pmgr->schedule_add (de->path, pfe (fh));
    assert (de->type == CFSREG);
    
    ptr<pfe> x = pmgr->lookup (get_parent_dir (de->path));
    if (x) {
      x->cr_addr = 0;
      pmgr->schedule_add (get_parent_dir (de->path), *x);
    }

    if (work == 0)
      done (cb, error);
  }

  void 
  checkin_reply (int i, cbi cb, bigint fh, int status)
  {
    work--;
    if (status) {
      warn << "commit: cannot checkin "
	   << up->to_add [i].path << " to repository\n";
      error = -1;
      if (work == 0)
        done (cb, error);
      return;
    }

    pde_t *de = &up->to_add [i];
    committed.push_back (de->path);
    committed_type.push_back ('A');
    pmgr->schedule_add (de->path, pfe (fh));

    // for commit notification message
    notify_cmds.push_back ("A");
    notify_files.push_back (de->path);
    
    ptr<pfe> x = pmgr->lookup (get_parent_dir (de->path));
    if (x) {
      x->cr_addr = 0;
      pmgr->schedule_add (get_parent_dir (de->path), *x);
    }

    if (de->type == CFSDIR)
      pmgr->schedule_mkdir (de->path, de->cfspath);

    if (work == 0)
      done (cb, error);
  }

  void do_add (cbi cb)
  {
    for (unsigned i=0; i<up->to_add.size (); i++) {
      pde_t e;
      e.name  = up->to_add [i].name;
      e.type  = up->to_add [i].type;
      e.path  = up->to_add [i].path;
      e.dirfh = up->to_add [i].dirfh;
      e.fh    = up->to_add [i].fh;
      work++;
      if (e.fh == 0)
	checkin (e, wrap (mkref (this), &commit_object::checkin_reply, i, cb));
      else
	relink (e, wrap (mkref (this), &commit_object::relink_reply, i, cb));
    }
  }

  void
  update_done (str path, cbi cb, int err)
  {
    if (err || up->has_conflict || up->requires_update) {
      if (err || up->has_conflict)
	warn << "Please resolve error or conflicts before committing\n";
      else if (up->requires_update)
	warn << "The repository has changed, please update before committing\n";
      // need to set err to non-zero so pastwatch will stop and cancel
      // the commit
      err = 1;
      cb (err);
    } else if (up->num_changes() < 1) {
      warn << "No changes, exiting without committing.\n";
      // This is a bit of a hack. It prevents pastwatch from
      // committing when there are no changes.  We should probably
      // have some way of telling driver.C so that it explicitly knows
      // not to insert the commit record, rather than toggling the
      // global variable <arg_readonly>.
      arg_readonly = true;  
      cb (-1);
    } else {
      warn << "committing in " << path << "\n";
      stage = 0;
      work++;
      do_remove (cb);
      work--;
      if (work == 0)
        done (cb, error);
    }
  }

public:
  commit_object (bigint va, bigint root, str path, cbi cb)
    : va (va), root (root), work (0), error (0)
  {
    warn << "checking for updates and conflicts\n";
    up = update_object_spec::alloc
      (root, path,
       wrap (mkref (this), &commit_object::update_done, path, cb), true, false);
  }
};

void
commit (bigint root, str path, cbi cb)
{
  ptr<commit_object> a = New refcounted<commit_object>
    (arg_projid, root, path, cb);
}

