/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "pastwatch.h"
#include "cfsrw.h"
#include "path.h"
#include "io.h"

extern ptr<cfsrw> _cfsrw;
extern bool command_done_called;

static void
noop_return (ptr<cfswcall> c)
{
  c->reply (CFSRWERR_INTR);
}

void
cfsw_call (ptr<cfswcall> c)
{
  if (command_done_called && c->proc () != CFSWPROC_COMMIT &&
      c->proc () != CFSWPROC_ABORT) {
    delaycb (0, wrap (noop_return, c));
    return;
  }
  _cfsrw->dispatch (c);
}

void
cfsr_call (bigint cr_addr, ptr<cfsrcall> c)
{
  c->cr_addr = cr_addr;
  _cfsrw->dispatch (c);
}

void
cfsr_call (ptr<cfsrcall> c)
{
  _cfsrw->dispatch (c);
}

typedef vec<bigint> bigintv;
typedef ptr<bigintv> bigintvp;
typedef vec<str> strv;
typedef ptr<strv> strvp;
typedef callback<void, bigintvp, cfsrw_status>::ref cblookupp_t;

static void next_lookup (bigint, strvp, cbcfsr);

void
cfsr_lookup_one (bigint dir, str name, cbcfsr cb)
{
  cfsrcallp c = New refcounted<cfsrcall> (CFSRPROC_LOOKUP, cb);
  cfsr_lookup_arg *arg = static_cast<cfsr_lookup_arg *> (c->arg ());
  arg->dir = dir;
  arg->name = name;
  cfsr_call (c);
}

static void
lookup_reply (strvp names, cbcfsr cb, cfsrcallp c, cfsrw_status status)
{
  if (status || names->size () == 0) {
    cb (c, status);
    return;
  }
  cfsr_lookup_res *res = static_cast<cfsr_lookup_res *> (c->res ());
  next_lookup (res->object, names, cb);
}

static void
next_lookup (bigint dir, strvp names, cbcfsr cb)
{
  str p = names->pop_front ();
  cfsr_lookup_one (dir, p, wrap (lookup_reply, names, cb));
}

void
cfsr_lookup (bigint dir, str path, cbcfsr cb)
{
  strvp names = New refcounted<strv>;
  splitpath (*names, path);
  assert (names->size ());
  next_lookup (dir, names, cb);
}

static void
next_lookup_v (bigint dir, strvp names, bigint s, cbcfsr cb);

static void
cfsr_lookup_one_v (bigint dir, str name, bigint s, cbcfsr cb)
{
  cfsrcallp c = New refcounted<cfsrcall> (CFSRPROC_LOOKUP, cb);
  cfsr_lookup_arg *arg = static_cast<cfsr_lookup_arg *> (c->arg ());
  arg->dir = dir;
  arg->name = name;
  cfsr_call (s, c);
}

static void
lookup_reply_v (strvp names, bigint s, cbcfsr cb,
                cfsrcallp c, cfsrw_status status)
{
  if (status || names->size () == 0) {
    cb (c, status);
    return;
  }
  cfsr_lookup_res *res = static_cast<cfsr_lookup_res *> (c->res ());
  next_lookup_v (res->object, names, s, cb);
}

static void
next_lookup_v (bigint dir, strvp names, bigint s, cbcfsr cb)
{
  str p = names->pop_front ();
  cfsr_lookup_one_v (dir, p, s, wrap (lookup_reply_v, names, s, cb));
}

void
cfsr_lookup_v (bigint dir, str path, bigint s, cbcfsr cb)
{
  strvp names = New refcounted<strv>;
  splitpath (*names, path);
  assert (names->size ());
  next_lookup_v (dir, names, s, cb);
}

static void
next_lookup_path (bigint fh, strvp names, bigintvp fhs, cblookupp_t cb);

static void
lookup_path_reply (strvp names, bigintvp fhs, cblookupp_t cb,
                   cfsrcallp c, cfsrw_status status)
{
  if (status) {
    cb (fhs, status);
    return;
  }
  
  cfsr_lookup_res *res = static_cast<cfsr_lookup_res *> (c->res ());
  fhs->push_back (res->object);

  if (names->size () == 0) {
    cb (fhs, status);
    return;
  }
  next_lookup_path (res->object, names, fhs, cb);
}

static void
next_lookup_path (bigint fh, strvp names, bigintvp fhs, cblookupp_t cb)
{
  str p = names->pop_front ();
  cfsr_lookup_one (fh, p, wrap (lookup_path_reply, names, fhs, cb));
}

void
cfsr_lookup_path (bigint dir, str path, cblookupp_t cb)
{
  bigintvp fhs = New refcounted<bigintv>;
  strvp names = New refcounted<strv>;
  splitpath (*names, path);
  assert (names->size ());
  next_lookup_path (dir, names, fhs, cb);
}


static void next_mkdir (bigint, strvp, cbcfsw);

void
cfsw_mkdir_one (bigint dir, str name, cbcfsw cb)
{
  cfswcallp c = New refcounted<cfswcall> (CFSWPROC_MKDIR, cb);
  cfsw_create_arg *arg = static_cast<cfsw_create_arg *> (c->arg ());
  arg->dir = dir;
  arg->name = name;
  arg->attributes.valid_size = 1;
  arg->attributes.size = 1024;
  arg->attributes.valid_mode = 1;
  arg->attributes.mode = 0755;
  cfsw_call (c);
}

static void
mkdir_reply (strvp names, cbcfsw cb, cfswcallp c, cfsrw_status status)
{
  if (status || names->size () == 0) {
    cb (c, status);
    return;
  }
  cfsw_create_res *res = static_cast<cfsw_create_res *> (c->res ());
  next_mkdir (res->object, names, cb);
}

static void
next_mkdir (bigint fh, strvp names, cbcfsw cb)
{
  str p = names->pop_front ();
  cfsw_mkdir_one (fh, p, wrap (mkdir_reply, names, cb));
}

void
cfsw_mkdir (bigint dir, str path, cbcfsw cb)
{
  strvp names = New refcounted<strv>;
  splitpath (*names, path);
  assert (names->size ());
  next_mkdir (dir, names, cb);
}

