/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <dirent.h>
#include "rxx.h"
#include "qhash.h"
#include "btree.h"
#include "stdinc.h"
#include "opts.h"
#include "path.h"
#include "metadata.h"

extern "C" {
time_t get_date (char *raw);
};

const char* cmd_list[] = {
  "import", "commit", "rename", "rtag", "branch",
  "history", "tags", "latest", "fetch", "checkout", "co",
  "merge", "update", "up", "add", "remove", "clear",
  "status", "diff", "log", "rescue", "sync"
};
const unsigned num_commnds = 21;

const int cmd_import = 0;
const int cmd_checkout = 1;
const int cmd_update = 2;
const int cmd_commit = 3;
const int cmd_add = 4;
const int cmd_remove = 5;
const int cmd_rename = 6;
const int cmd_clear = 7;
const int cmd_diff = 8;
const int cmd_status = 9;
const int cmd_log = 10;
const int cmd_rescue = 11;
const int cmd_branch = 12;
const int cmd_rtag = 13;
const int cmd_fetch = 14;
const int cmd_merge = 15;
const int cmd_latest = 16;
const int cmd_history = 17;
const int cmd_tags = 18;
const int cmd_sync = 19;

str arg_localpath = "";
str arg_branchtag = "";
str arg_jointag = "";
str arg_branchfrom = "";
str arg_projstr;
str arg_logkeyfile;
str arg_userkeyfile;
str arg_keystr;
str arg_branch;
str arg_message = "";
str arg_tag[2] = {"", ""};
str arg_date[2] = {"", ""};
str arg_timestamp = "";
str arg_newtag = "";

int arg_command = -1;
int arg_debugmode = 0;
int arg_revision[2] = {0,0};

bool arg_noupdate = false;
bool arg_resetsticky = false;
bool arg_readonly = true;
bool arg_updatedir = false;
bool arg_checkonly = false;
bool arg_force = false;
bool arg_contextdiff = false;
bool arg_relink = false;
bool arg_dironly = false;
bool arg_recurse = true;
bool arg_killbranch = false;
bool arg_newbranch = false;
bool arg_notrans = false;
bool arg_ignore_implicit_branches = false;
bool arg_sync_push = false;

bigint arg_projid;
bigint arg_key;
bigint arg_handle;

vec<str> arg_projids;
vec<str> arg_userkeys;
vec<str> arg_logkeys;
vec<str> arg_branches;
vec<ptr<vec<str> > > arg_paths;

extern int  check_dirent (str dirpath, dirent *e, bool& ignore_entry);
extern void check_skip   (str dirpath, str name, bool& ignore_entry);
extern void list_branches ();
extern void clean_and_exit (int error);


#define ATOI(s,v) \
{ \
  char *end = 0; \
  v = strtol (s, &end, 10); \
  if (*end != '\0') { \
    usage (); \
    return 1; \
  } \
}

bool
is_cmd(char *s) {
  for(unsigned i=0; i<num_commnds; i++)
    if (!strcmp(s, cmd_list[i]))
      return true;
  return false;
}

str
get_btag_from_infofile (str path)
{
  str x;
  str p = get_info_file (get_parent_dir (path));
  str buf = file2str (p);
  if (!buf) {
    p = get_info_file (get_dir (path));
    buf = file2str (p);
  }
  if (buf) {
    vec<str> n;
    split (&n, "\n", buf);
    assert (n.size () >= 4);
    if (n.size () == 4)
      return x;
    else
      return n [4];
  }
  return x;
}

str
get_commit_record_from_infofile (str path)
{
  str p = get_info_file (get_parent_dir (path));
  str buf = file2str (p);
  if (!buf) {
    p = get_info_file (get_dir (path));
    buf = file2str (p);
  }
  if (buf) {
    vec<str> n;
    split (&n, "\n", buf);

    if (n.size () < 4) {
      warn << "Error in Pastwatch/info file.  File does not contain branch field.\n";
      assert(0);
    } 
    return n [3];
  }
  return NULL;
}

static str
check_projectid (str projlnk) {
  // Tries to return armor64A of the project.id from the symbolic link.
  // Returns "" if it can't get the project id.
  struct stat sb;
  int err = lstat(projlnk.cstr(), &sb);
  if (!err && (sb.st_mode & S_IFLNK) == S_IFLNK) {
    str n;
    char sym[1024];
    rxx rx ("project:([0-9a-zA-Z+-]*)[/]*$");
    int r = readlink (projlnk.cstr (), sym, 1024);
    if (r > 0) {
      sym[r] = '\0';
      if (rx.search(sym))
	return rx [1];
      return "";
    }
  }
  return "";
}

static rxx logkey_rxx ("^logkey\\.[0-9a-zA-Z+-]+$");
static int
find_logkey (str projid, char *projlabel, char *sk_file, str &chosen_key) {
  // Sets <chosen_key>. <projid> is encoded in armor64A

  if (sk_file && strcmp(sk_file, "")) {
    chosen_key = str(sk_file);
    return 0;
  }

  str replicadir = pget_replicadir() << "/" << projid;

  struct stat sb;
  DIR *dir;
  int err = lstat(replicadir.cstr(), &sb);

  if (!err && (sb.st_mode & S_IFDIR) == S_IFDIR && 
      (dir = opendir (replicadir.cstr()))) {

    vec<str> potential_keys;
    struct dirent *e;

    while ((e = readdir (dir)))
      if (logkey_rxx.search(str(e->d_name)))
	potential_keys.push_back(str(e->d_name));
    closedir (dir);

    if (potential_keys.size() > 1) {
      warn << "Multiple member keys for " << projlabel
	   << ".  You must specify one with the -s option.\n";
      for(unsigned i=0; i<potential_keys.size(); i++)
	warn << "  Key: " << replicadir << "/" << potential_keys[i] << "\n";
      clean_and_exit (1);
    } else if (potential_keys.size() == 1) {
      chosen_key = replicadir << "/" << potential_keys[0];
      return 0;
    }
  }

  chosen_key = str("");
  return 0;
}

static rxx memberkey_rxx ("^member\\.[0-9a-zA-Z+-]+$");
static int
find_memberkey (str projlabel, str &memberkey) {
  // Returns 0 if found and sets <memberkey> to the file location
  // of the member's secret key.

  if (!projlabel)
    return -1;

  str projdir = pget_pwroot() << "/" << projlabel;
  struct stat sb;
  DIR *dir;
  int err = lstat(projdir.cstr(), &sb);

  if (!err && (sb.st_mode & S_IFDIR) == S_IFDIR && 
      (dir = opendir (projdir.cstr()))) {

    vec<str> potential_keys;
    struct dirent *e;

    while ((e = readdir (dir)))
      if (memberkey_rxx.search(str(e->d_name)))
	potential_keys.push_back(str(e->d_name));
    closedir (dir);

    if (potential_keys.size() > 1) {
      warn << "Multiple member keys for " << projlabel
	   << ".  You must specify one with the -u option.\n";
      for(unsigned i=0; i<potential_keys.size(); i++)
	warn << "  Key: " << projdir << "/" << potential_keys[i] << "\n";
      clean_and_exit (1);
    } else if (potential_keys.size() == 1) {
      memberkey = projdir << "/" << potential_keys[0];
      return 0;
    }
  }
  
  return -1;
}

static int
get_info (str path, char *sk_file, char *projstr, char *branch)
{
  // get_info finds info for each file passed to pastwatch including:
  //   a) member key
  //   b) project block
  //   c) branch
  //   d) user key (only if pastwatch needs to create a new log for this user)
  // It appends this items to the vectors read by driver.C 

  str lkstr = "", pstr = "", bstr = "", mstr = "";
  ptr<info_file> info = New refcounted<info_file> ();
  int r = -1;
  if (arg_command != cmd_import && 
      arg_command != cmd_checkout)
    r = info->read_file (path); // Only read info file if exists.

  // Figure out project id
  bigint pid;
  if (pstr == "" && projstr)
    pstr = check_projectid (str(projstr));
  if (pstr == "" && projstr) {
    str p = pget_pwroot() << "/" << projstr << "/project.id";
    pstr = check_projectid (p);
  }
  if (pstr == "" && !r)
    pstr = armor64A(info->get_projid());
  if (pstr == "" && dearmor64A_bigint (projstr, pid))
    pstr = projstr;
  if (pstr == "")
    return -1;
  
  // Figure out the log key
  find_logkey(pstr, projstr, sk_file, lkstr);
  if (lkstr == "" && !r)
    lkstr = info->get_logkey (); // If not found, try the info file.

  // Look for the member key.
  str projlabel = projstr;
  if (!projlabel || projlabel == "")
    projlabel = info->get_projlabel();
  str memberkey = "";
  if (!find_memberkey(projlabel, memberkey))
    mstr = memberkey;

  // Figure out branch
  if (branch)
    bstr = str (branch, strlen (branch));
  if ( bstr == "" && !r)
    bstr = info->get_branchlabel ();

  if (pstr == "") {
    warn << "Could not determine project id\n";
    return -1;
  }

  // If the (project, key, and branch) are all the same, only add a
  // new path to the entry.
  for (unsigned i=0; i<arg_projids.size (); i++) {
    if (arg_projids [i] == pstr &&
	arg_logkeys [i] == lkstr &&
	arg_userkeys [i] == mstr &&
	arg_branches [i] == bstr) {
      ptr<vec<str> > z = arg_paths [i];
      z->push_back (path);
      arg_paths [i] = z;
      return 0;
    }
  }

  // Create a new (project, key, and branch) for this entry.
  arg_projids.push_back (pstr);
  arg_logkeys.push_back (lkstr);
  arg_userkeys.push_back (mstr);
  arg_branches.push_back (bstr);
  ptr<vec<str> > z = New refcounted<vec<str> >;
  z->push_back (path);
  arg_paths.push_back (z);
  return 0;
}

static void
usage ()
{
  warnx << "Pastwatch version " << VERSION << "\n";
  warnx << "usage: " << progname << " [opts] cmd [cmd opts] files\n";
  warnx << " -n: check only, don't make changes\n";
  warnx << " -l: don't open sub-directories\n";
  warnx << " -b <branch tag>: specify a branch to use\n";
  warnx << " -i: ignore implicit branches when possible\n";
  warnx << " -x <n>: debug mask\n";
  warnx << "   1 - reports latency to fetch project block and public keys\n";
  warnx << "   2 - reports latency to fetch blocks\n";
  warnx << "   4 - outputs debug messages when updating a directory\n";
  warnx << " -s <file>: public/private key file\n";
  warnx << " -p <project id>: project identification number\n";
  warnx << " See the man page for commands and command options: pastwatch(1).\n";
}

int update_debug = 0;
int ns_debug = 0;
int cfs_debug = 0;
int projinf_debug = 0;

int
getopts (int argc, char **argv)
{
  char *sk_file = 0;
  char *projstr = 0;
  char *branch = 0;
  int    pastargc = 0;

  if (argc == 1) {
    usage ();
    return 1;
  }

  for(int i=0; i<argc; i++) {
    if (is_cmd(argv[i])) {
      pastargc = i;
      break;
    }
  }

  int ch;
  while ((ch = getopt (pastargc, argv, "nlb:x:s:p:i")) != -1) {
    switch (ch) {
    case 'b':
      branch = optarg;
      break;
    case 'x':
      ATOI (optarg, arg_debugmode);
      break;
    case 'l':
      arg_recurse = false;
      break;
    case 'n':
      arg_checkonly = true;
      break;
    case 'p':
      projstr = optarg;
      break;
    case 's':
      sk_file = optarg;
      break;
    case 'i':
      arg_ignore_implicit_branches = true;
      break;
    case '?':
    default:
      usage ();
      return 1;
    }
  }
  argc -= optind;
  argv += optind;

  if (argc == 0) {
    usage ();
    return 1;
  }

  if (arg_debugmode & 1)
    projinf_debug = 1;
  if (arg_debugmode & 2)
    ns_debug = 1;
  if (arg_debugmode & 4)
    update_debug = 1;

  if (strcmp (argv[0], "import") == 0) {
    arg_readonly = false;
    arg_command = cmd_import;
  }
  else if (strcmp (argv[0], "commit") == 0) {
    arg_readonly = false;
    arg_command = cmd_commit;
  }
  else if (strcmp (argv[0], "rename") == 0) {
    arg_readonly = false;
    arg_command = cmd_rename;
  }
  else if (strcmp (argv[0], "rtag") == 0) {
    arg_checkonly = true;
    arg_readonly = false;
    arg_command = cmd_rtag;
  }
  else if (strcmp (argv[0], "branch") == 0) {
    arg_checkonly = true;
    arg_readonly = false;
    arg_command = cmd_branch;
  }
  else if (strcmp (argv[0], "history") == 0) {
    arg_checkonly = true;
    arg_command = cmd_history;
  }
  else if (strcmp (argv[0], "tags") == 0) {
    arg_checkonly = true;
    arg_command = cmd_tags;
  }
  else if (strcmp (argv[0], "latest") == 0) {
    arg_checkonly = true;
    arg_command = cmd_latest;
  }
  else if (strcmp (argv[0], "fetch") == 0) {
    arg_checkonly = true;
    arg_command = cmd_fetch;
  }

  else if (strcmp (argv[0], "checkout") == 0)
    arg_command = cmd_checkout;

  else if (strcmp (argv[0], "co") == 0)
    arg_command = cmd_checkout;
  
  else if (strcmp (argv[0], "merge") == 0)
    arg_command = cmd_merge;

  else if (strcmp (argv[0], "update") == 0)
    arg_command = cmd_update;

  else if (strcmp (argv[0], "up") == 0)
    arg_command = cmd_update;

  else if (strcmp (argv[0], "add") == 0)
    arg_command = cmd_add;

  else if (strcmp (argv[0], "remove") == 0)
    arg_command = cmd_remove;

  else if (strcmp (argv[0], "clear") == 0)
    arg_command = cmd_clear;

  else if (strcmp (argv[0], "status") == 0)
    arg_command = cmd_status;

  else if (strcmp (argv[0], "diff") == 0)
    arg_command = cmd_diff;

  else if (strcmp (argv[0], "log") == 0)
    arg_command = cmd_log;

  else if (strcmp (argv[0], "rescue") == 0)
    arg_command = cmd_rescue;

  else if (strcmp (argv[0], "sync") == 0)
    arg_command = cmd_sync;

  // optreset = 1;
  optind = 1;

  //
  // add -S timestamp
  //
  if (arg_command == cmd_add) {
    while ((ch = getopt (argc, argv, "S:")) != -1)
      switch (ch) {
      case 'S':
	arg_timestamp = str (optarg, strlen (optarg));
        break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // remove -f -d
  //
  else if (arg_command == cmd_remove) {
    while ((ch = getopt (argc, argv, "df")) != -1)
      switch (ch) {
      case 'd':
	arg_dironly = true;
	break;
      case 'f':
        arg_force = true;
        break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // merge -d -t Tag
  //
  else if (arg_command == cmd_merge) {
    arg_noupdate = true;
    while ((ch = getopt (argc, argv, "dt:")) != -1)
      switch (ch) {
      case 'd':
        arg_updatedir = true;
        break;
      case 't':
	arg_tag[0] = str (optarg, strlen (optarg));
	break;
      default:
	usage ();
	return 1;
      }
    if (arg_tag[0] == "") {
      warn << "merge command requires a version\n";
      usage ();
      return 1;
    }
    argc -= optind;
    argv += optind;
  }
  //
  // update -A -d -D Date -r Rev -t Tag
  //
  else if (arg_command == cmd_update) {
    while ((ch = getopt (argc, argv, "AdD:r:t:")) != -1)
      switch (ch) {
      case 'A':
	arg_resetsticky = true;
	break;
      case 'd':
        arg_updatedir = true;
        break;
      case 'D':
	arg_date[0] = str (optarg, strlen (optarg));
	break;
      case 't':
	arg_tag[0] = str (optarg, strlen (optarg));
	break;
      case 'r':
	ATOI (optarg, arg_revision[0]);
	break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // checkout -t Tag -d Dir
  //
  else if (arg_command == cmd_checkout) {
    while ((ch = getopt (argc, argv, "t:d:")) != -1)
      switch (ch) {
      case 't':
	arg_tag[0] = str (optarg, strlen (optarg));
	break;
      case 'd':
        arg_localpath = str (optarg, strlen (optarg));
        break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // rtag -S timestamp -D date -t Tag Tag
  //
  else if (arg_command == cmd_rtag) {
    while ((ch = getopt (argc, argv, "S:D:t:")) != -1)
      switch (ch) {
      case 'S':
	arg_timestamp = str (optarg, strlen (optarg));
	break;
      case 'D':
	arg_date[0] = str (optarg, strlen (optarg));
	break;
      case 't':
	arg_tag[0] = str (optarg, strlen (optarg));
	break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // diff -c -D Date -t Tag -r Rev
  //
  else if (arg_command == cmd_diff) {
    while ((ch = getopt (argc, argv, "D:cr:t:")) != -1)
      switch (ch) {
      case 'c':
        arg_contextdiff = true;
        break;

      case 'D':
	if (arg_date[0] == "" && arg_tag[0] == "" && arg_revision[0] == 0) 
	  arg_date[0] = str (optarg, strlen (optarg));
	else
	  arg_date[1] = str (optarg, strlen (optarg));
	break;

      case 't':
	if (arg_date[0] == "" && arg_tag[0] == "" && arg_revision[0] == 0) 
	  arg_tag[0] = str (optarg, strlen (optarg));
	else
	  arg_tag[1] = str (optarg, strlen (optarg));
	break;

      case 'r':
	if (arg_date[0] == "" && arg_tag[0] == "" && arg_revision[0] == 0) 
	  { ATOI (optarg, arg_revision[0]); }
	else
	  { ATOI (optarg, arg_revision[1]); }
	break;

      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // branch -S timestamp -k -B -f tag -b str -m Msg
  //
  else if (arg_command == cmd_branch) {
    bool has_msg = false;
    while ((ch = getopt (argc, argv, "S:kBf:b:m:")) != -1)
      switch (ch) {
      case 'S':
	arg_timestamp = str (optarg, strlen (optarg));
	break;
      case 'B':
	arg_newbranch = true;
	break;
      case 'k':
	arg_killbranch = true;
	break;
      case 'f':
	arg_newbranch = true;
	arg_branchfrom = str (optarg, strlen (optarg));
	break;
      case 'b':
	arg_branchtag = str (optarg, strlen (optarg));
	break;
      case 'm':
	arg_message = str (optarg, strlen (optarg));
	has_msg = true;
        break;
      default:
	usage ();
	return 1;
      }
    if (arg_newbranch && arg_killbranch) {
      warn << "-k and -B are incompatible options\n";
      usage ();
      return 1;
    }
    if (arg_branchtag && arg_branchtag != "" && arg_killbranch) {
      warn << "-k and -b are incompatible options\n";
      usage ();
      return 1;
    }
    if (arg_branchfrom && arg_branchfrom != "" && arg_killbranch) {
      warn << "-f and -k are incompatible options\n";
      usage ();
      return 1;
    }
    if (!has_msg) {
      warn << "must provide a commit message using -m\n";
      usage ();
      return 1;
    }
    argc -= optind;
    argv += optind;
  }
  //
  // commit -S timestamp -j tag -m Msg
  //
  else if (arg_command == cmd_commit) {
    bool has_msg = false;
    while ((ch = getopt (argc, argv, "S:m:")) != -1)
      switch (ch) {
      case 'S':
	arg_timestamp = str (optarg, strlen (optarg));
	break;
      case 'j':
	arg_jointag = str (optarg, strlen (optarg));
	break;
      case 'm':
	arg_message = str (optarg, strlen (optarg));
	has_msg = true;
        break;
      default:
	usage ();
	return 1;
      }
    if (!has_msg) {
      warn << "must provide a commit message using -m\n";
      usage ();
      return 1;
    }
    argc -= optind;
    argv += optind;
  }
  //
  // rename -S timestamp -m Msg
  //
  else if (arg_command == cmd_rename) {
    bool has_msg = false;
    while ((ch = getopt (argc, argv, "S:m:")) != -1)
      switch (ch) {
      case 'S':
	arg_timestamp = str (optarg, strlen (optarg));
	break;
      case 'm':
	arg_message = str (optarg, strlen (optarg));
	has_msg = true;
        break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // import -S timestamp -m Msg
  //
  else if (arg_command == cmd_import) {
    bool has_msg = false;
    while ((ch = getopt (argc, argv, "S:m:")) != -1) {
      switch (ch) {
      case 'S':
	arg_timestamp = str (optarg, strlen (optarg));
	break;
      case 'm':
	arg_message = str (optarg, strlen (optarg));
	has_msg = true;
        break;
      default:
	usage ();
	return 1;
      }
    }
    if (!has_msg) {
      warn << "must provide a commit message using -m\n";
      usage ();
      return 1;
    }
    argc -= optind;
    argv += optind;
  }
  //
  // rescue -a -h Handle
  //
  else if (arg_command == cmd_rescue) {
    bool has_handle = false;
    while ((ch = getopt (argc, argv, "ah:")) != -1)
      switch (ch) {
      case 'a':
	arg_relink = true;
	break;
      case 'h':
	{
          str x = dearmor64A (str (optarg, strlen (optarg)));
          mpz_set_rawmag_be (&arg_handle, x.cstr(), x.len());
	  has_handle = true;
	}
        break;
      default:
	usage ();
	return 1;
      }
    if (!has_handle) {
      warn << "must specify a handle to rescue using -h\n";
      usage ();
      return 1;
    }
    argc -= optind;
    argv += optind;
  }
  //
  // log -d
  //
  else if (arg_command == cmd_log) {
    while ((ch = getopt (argc, argv, "d")) != -1)
      switch (ch) {
      case 'd':
	arg_dironly = true;
	break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // status -d
  //
  else if (arg_command == cmd_status) {
    while ((ch = getopt (argc, argv, "d")) != -1)
      switch (ch) {
      case 'd':
	arg_dironly = true;
	break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // sync
  //
  else if (arg_command == cmd_sync) {
    if (optind < argc && !strcmp(argv[optind], "push")) {
      arg_sync_push = true;
      argc--;
      argv++;
    }
    if (optind < argc) {
      usage ();
      return 1;
    }
  }
  
  else {
    argc -= 1;
    argv += 1;
  }

  //  Check arguments -------------------------------------------------

  if (arg_command == -1) {
    usage ();
    return 1;
  }

  if (arg_revision[0] != 0 && arg_tag[0] && arg_tag[0] != "") {
    warn << "cannot accept both a revision and a tag\n";
    usage ();
    return 1;
  }

  if (arg_date[0] && arg_date[0] != "" && arg_revision[0] != 0) {
    warn << "cannot accept both a revision and a date\n";
    usage ();
    return 1;
  }
  
  if (arg_date[0] && arg_date[0] != "" && arg_tag[0] && arg_tag[0] != "") {
    warn << "cannot accept both a tag and a date\n";
    usage ();
    return 1;
  }

  if (arg_date[0] && arg_date[0] != "") {
    time_t t = get_date ((char *)arg_date[0].cstr ());
    if (t < 0) {
      warn << "cannot parse " << arg_date[0] << "\n";
      usage ();
      return 1;
    }
  }

  vec<str> paths;

  if (argc == 0) {
    if (arg_command == cmd_import || arg_command == cmd_checkout ||
	arg_command == cmd_add || arg_command == cmd_rescue) {
      warn << "must provide a file or directory name\n";
      usage ();
      return 1;
    }
    else if (arg_command == cmd_rtag) {
      warn << "must provide a tag\n";
      usage ();
      return 1;
    }
    paths.push_back (".");
  }
  else if (argc > 1) {
    if (arg_command == cmd_import || arg_command == cmd_checkout ||
	arg_command == cmd_rescue) {
      warn << "please give only one file name\n";
      usage ();
      return 1;
    }
    else if (arg_command == cmd_rtag) {
      warn << "rtag command only takes one argument\n";
      usage ();
      return 1;
    }
  }
  else if (argc == 1) {
    str x = sanitize_path (argv [0]);
    if (arg_command == cmd_import && (x == "." || x == "..")) {
      warn << "import - invalid module name\n";
      return 1;
    }
    if (arg_command == cmd_checkout && (x == "/")) {
      warn << "checkout - invalid module name\n";
      return 1;
    }
  }

  if (arg_command == cmd_import || arg_command == cmd_checkout) {
    if (!projstr) {
      warn << "no project id\n";
      usage ();
      return 1;
    }
  }

  //---------------------------------------------------------

  for (int i=0; i<argc; i++) {
    str x = sanitize_path (argv [i]);
    paths.push_back (x);
  }

  while (!paths.empty ()) {
    str path = paths.pop_front ();
    struct stat sb;
    int err = lstat (path, &sb);
  
    int r = get_info (path, sk_file, projstr, branch);
    if (r) {
      //warn << "Error getting info (1)\n";
      return r;
    }

    // these commands do not need files, so don't recurse in the
    // current directory. if can't find a memberlist, user block or key, just stop.
    else if (r < 0 && (arg_command == cmd_branch ||
	               arg_command == cmd_rtag ||
	               arg_command == cmd_fetch ||
	               arg_command == cmd_latest ||
	               arg_command == cmd_tags ||
		       arg_command == cmd_history)) {
      warn << "no key or project id\n";
      return 1;
    }

    // r < 0 iff the current directory does not contain a Pastwatch
    // directory. in this case we recurse.
    else if (r < 0 && !err && (sb.st_mode & S_IFDIR)) {
      DIR *dir = opendir (path);
      if (!dir)
        warn << "cannot open " << path << "\n";
      else {
        struct dirent *e; 
        while ((e = readdir (dir))) {
          bool blah = false;
          if (check_dirent (path, e, blah))
	    continue;
	  bool skip = false;
	  check_skip (path, str (e->d_name), skip);
	  if (skip)
	    continue;
          str p = path << "/" << e->d_name;
          int r2 = get_info (p, sk_file, projstr, branch);
	  if (r2) {
	    //warn << "Error getting info (2)\n";
	    return r2;
	  }
        }
        closedir (dir);
      }
    }
    else if (r != 0) {
      warn << "no key or project id for " << path << "\n";
      return 1;
    }
  }
  
  if (arg_command == cmd_rename) {
    if (arg_projids.size () != 1) {
      warn << "rename command needs two files from the same project\n";
      usage ();
      return 1;
    }
    if (arg_paths [0]->size () != 2) {
      warn << "rename command needs two files from the same project\n";
      usage ();
      return 1;
    }
  }
  return 0;
}  

str
cmd2str (int cmd)
{
  str s;
  if (cmd == cmd_checkout)
    s = "checkout";
  else if (cmd == cmd_merge)
    s = "merge";
  else if (cmd == cmd_update)
    s = "update";
  else if (cmd == cmd_import)
    s = "import";
  else if (cmd == cmd_commit)
    s = "commit";
  else if (cmd == cmd_add)
    s = "add";
  else if (cmd == cmd_remove)
    s = "remove";
  else if (cmd == cmd_clear)
    s = "clear";
  else if (cmd == cmd_log)
    s = "log";
  else if (cmd == cmd_status)
    s = "status";
  else if (cmd == cmd_diff)
    s = "diff";
  else if (cmd == cmd_rescue)
    s = "rescue";
  else if (cmd == cmd_rename)
    s = "rename";
  else if (cmd == cmd_tags)
    s = "tags";
  else if (cmd == cmd_history)
    s = "history";
  else if (cmd == cmd_latest)
    s = "latest";
  else if (cmd == cmd_fetch)
    s = "fetch";
  else if (cmd == cmd_branch)
    s = "branch";
  else if (cmd == cmd_rtag)
    s = "rtag";
  else
    assert (0);
  return s;
}

